import pyhepmc as hep
import numpy as np
import pytest


def test_no_vertex_info():
    px = py = pz = en = m = np.linspace(0, 1, 4)

    pid = np.arange(4) + 1
    sta = np.zeros(4, dtype=np.int32)
    parents = [(0, 0), (1, 1), (0, 0), (0, 0)]
    hev = hep.GenEvent()
    hev.from_hepevt(0, px, py, pz, en, m, pid, sta, parents)
    assert len(hev.vertices) == 1
    assert len(hev.particles) == 4


def test_parents_range_exceeding_particle_range():
    px = py = pz = en = m = np.linspace(0, 1, 6)
    pid = np.arange(6) + 1
    sta = np.zeros(6, dtype=np.int32)
    parents = [(0, 0), (1, 1), (2, 0), (3, 5), (4, 10), (3, 5)]
    with pytest.raises(RuntimeError):
        hep.GenEvent().from_hepevt(0, px, py, pz, en, m, pid, sta, parents)


def test_invalid_length_of_parents():
    px = py = pz = en = m = np.linspace(0, 1, 3)
    pid = np.arange(3) + 1
    sta = np.zeros(3, dtype=np.int32)
    parents = [(0, 0), (1, 2)]
    with pytest.raises(RuntimeError):
        hep.GenEvent().from_hepevt(0, px, py, pz, en, m, pid, sta, parents)


def test_inverted_parents_range():
    px = py = pz = en = m = vx = vy = vz = vt = np.linspace(0, 1, 4)
    pid = np.arange(4) + 1
    sta = np.zeros(4, dtype=np.int32)
    # inverted range is not an error (2, 1) will be converted to (1, 2)
    parents = [(0, 0), (2, 1), (3, 3), (3, 3)]
    hev = hep.GenEvent()
    hev.from_hepevt(0, px, py, pz, en, m, pid, sta, parents)
    expected = [[0, 1], [2]]
    got = [[p.id - 1 for p in v.particles_in] for v in hev.vertices]
    assert expected == got


@pytest.mark.parametrize("bad", ([-4, 1], [1, -4]))
def test_negative_parents_range(bad):
    px = py = pz = en = m = vx = vy = vz = vt = np.linspace(0, 1, 4)
    pid = np.arange(4) + 1
    sta = np.zeros(4, dtype=np.int32)
    # inverted range is not an error (2, 1) will be converted to (1, 2)
    parents = [(0, 0), bad, (3, 3), (3, 3)]
    with pytest.raises(RuntimeError):
        hep.GenEvent().from_hepevt(0, px, py, pz, en, m, pid, sta, parents)



from pyhepmc._core import (
    _Setup_print_errors,
    _Setup_set_print_errors,
    _Setup_print_warnings,
    _Setup_set_print_warnings,
    _Setup_debug_level,
    _Setup_set_debug_level,
)
from typing import Any


class _SetupMeta(type):
    @property
    def print_errors(cls) -> bool:
        "Whether to print errors or not."
        return _Setup_print_errors()  # type:ignore

    @property
    def print_warnings(cls) -> bool:
        "Whether to print warnings or not."
        return _Setup_print_warnings()  # type:ignore

    @property
    def debug_level(cls) -> int:
        "Access debug level."
        return _Setup_debug_level()  # type:ignore

    def __setattr__(self, name: str, value: Any) -> None:
        attr = {
            "print_errors": _Setup_set_print_errors,
            "print_warnings": _Setup_set_print_warnings,
            "debug_level": _Setup_set_debug_level,
        }
        fn = attr.get(name, None)
        if fn is None:
            raise AttributeError
        fn(value)


class Setup(metaclass=_SetupMeta):
    """
    Imitates the Setup namespace.

    You can directly read and write to the attributes of this class
    without creating an instance. They manipulate the corresponding
    global values in the HepMC3 C++ library.
    """

    __slots__ = ("print_errors", "print_warnings", "debug_level")



import pyhepmc
import pyhepmc.io
import pyhepmc._core as core


def test_init():
    core_names = set()

    for name in dir(core):
        if name.startswith("_"):
            continue
        core_names.add(name)

    core_names.remove("stringstream")
    core_names.remove("iostream")
    core_names.remove("pyiostream")
    core_names.remove("Attribute")
    core_names.remove("AttributesView")
    core_names.remove("Reader")
    core_names.remove("Writer")
    core_names.remove("RunInfoAttributesView")
    core_names.remove("NumpyAPI")
    core_names.remove("ParticlesAPI")
    core_names.remove("VerticesAPI")

    names = set()
    for name in dir(pyhepmc):
        if name.startswith("_"):
            continue
        names.add(name)
    for name in dir(pyhepmc.io):
        if name.startswith("_"):
            continue
        names.add(name)

    assert core_names - names == set()



# import pytest
import pyhepmc as hep


def test_genruninfo():
    ri = hep.GenRunInfo()
    ri.tools = [("foo", "1.0", "bar")]
    ri.weight_names = ["a", "b", "c"]
    assert repr(ri) == (
        "GenRunInfo(tools=[ToolInfo(name='foo', version='1.0', description='bar')], "
        "weight_names=['a', 'b', 'c'], attributes={})"
    )



import dataclasses
from typing import Union, Dict, Type
import subprocess as subp


class Value(str):
    pass


class Quoted(str):
    def __str__(self) -> str:
        s = super().__str__()
        return f'"{s}"'


class Block(Dict[str, Union[Quoted, Value]]):
    def __init__(self, **kwargs: str):
        super().__init__()
        for k, v in kwargs.items():
            self.__setitem__(k, v)

    def __setitem__(self, key: str, value: str) -> None:
        if not value:
            return
        type: Union[Type[Value], Type[Quoted]] = Value
        if key in ("size", "label", "tooltip", "labeltooltip"):
            type = Quoted
        super().__setitem__(key, type(value))

    def __str__(self) -> str:
        s = []
        for k, v in self.items():
            s.append(f"{k}={v}")
        return f"[{', '.join(s)}]"

    def __repr__(self) -> str:
        s = []
        for k, v in self.items():
            s.append(f"{k}={v!r}")
        return f"Block({', '.join(s)})"


@dataclasses.dataclass
class Digraph:
    name: Quoted
    graph_attr: Block
    node_attr: Block
    nodes: Dict[str, Block]
    edges: Dict[str, Block]

    def __init__(
        self,
        *,
        name: str = "",
        graph_attr: Block = None,
        node_attr: Block = None,
        nodes: Dict[str, Block] = None,
        edges: Dict[str, Block] = None,
    ):
        self.name = Quoted(name)
        self.graph_attr = graph_attr or Block()
        self.node_attr = node_attr or Block()
        self.nodes = nodes or {}
        self.edges = edges or {}

    def edge(
        self,
        vin: str,
        vout: str,
        *,
        dir: str = "none",
        label: str = "",
        color: str = "",
        style: str = "",
        penwidth: str = "",
        tooltip: str = "",
        labeltooltip: str = "",
    ) -> None:
        self.edges[f"{vin} -> {vout}"] = Block(
            dir=dir,
            label=label,
            color=color,
            style=style,
            penwidth=penwidth,
            tooltip=tooltip,
            labeltooltip=labeltooltip,
        )

    def node(self, id: str, *, style: str = "", tooltip: str = "") -> None:
        self.nodes[f"{id}"] = Block(tooltip=tooltip, style=style)

    def __str__(self) -> str:
        s = [
            f"graph {self.graph_attr}",
            f"node {self.node_attr}",
        ]
        for k, v in self.nodes.items():
            s.append(f"{k} {v}")
        for k, v in self.edges.items():
            s.append(f"{k} {v}")
        nl = "\n"
        return f"digraph {self.name} {{\n{nl.join(s)}\n}}"

    def pipe(self, *, format: str, encoding: str = None) -> Union[str, bytes]:
        input = str(self)
        try:
            r = subp.run(
                ["dot", f"-T{format}"], capture_output=True, input=input.encode("utf-8")
            )
        except FileNotFoundError:
            raise FileNotFoundError(
                "command `dot` cannot be found, please install graphviz"
            )
        if r.returncode != 0:
            msg = r.stderr.decode("utf-8")
            match = "Error: <stdin>: "
            idx = msg.find(match)
            if idx >= 0:
                idx += len(match)
                msg = msg[idx:]
            msg += "\n" + input
            raise ValueError(msg)
        if encoding:
            return r.stdout.decode(encoding)
        return r.stdout

    def _repr_svg_(self) -> str:
        s = self.pipe(format="svg", encoding="utf-8")
        assert isinstance(s, str)
        return s

    def _repr_png_(self) -> bytes:
        b = self.pipe(format="png")
        assert isinstance(b, bytes)
        return b

    def _repr_html_(self) -> str:
        return self._repr_svg_()



import pytest
from test_basic import make_evt
import pyhepmc
from pyhepmc import view
from pathlib import Path
import io
import os
import numpy as np

CDIR = Path(__file__).parent
RESULT_DIR = CDIR / "fig"
REFERENCE_DIR = CDIR / "data"
RESULT_DIR.mkdir(exist_ok=True)

DOT_IS_AVAILABLE = bool(view.SUPPORTED_FORMATS)
PARTICLE_IS_AVAILABLE = True
try:
    import particle  # noqa
except ModuleNotFoundError:
    PARTICLE_IS_AVAILABLE = False


@pytest.fixture
def evt():
    return make_evt()


@pytest.fixture
def graph():
    return view.to_dot(make_evt())


def test_to_dot_1(graph):
    s = str(graph)
    assert s.startswith('digraph "dummy-1.0\nevent number = 1')
    assert "in_1" in s
    assert "in_2" in s
    assert "in_3" not in s
    assert "out_6" in s
    assert "out_7" in s
    assert "out_8" in s


def test_to_dot_2():
    ev = pyhepmc.GenEvent()
    # unknown particle
    p = pyhepmc.GenParticle((0, 0, 0, 1e-3), pid=91, status=1)
    ev.add_particle(p)
    # invalid particle
    q = pyhepmc.GenParticle((0, 0, 0, 1e-3), pid=0, status=1)
    ev.add_particle(q)
    v = pyhepmc.GenVertex()
    v.add_particle_in(p)
    v.add_particle_out(q)
    ev.add_vertex(v)
    d = view.to_dot(ev)
    s = str(d)
    if PARTICLE_IS_AVAILABLE:
        assert "Invalid(0)" in s
        assert "Internal(91)" in s
    else:
        assert "PDGID(0)" in s


def test_to_dot_3(evt):
    d = view.to_dot(evt, size=(5, 6))
    assert d.graph_attr["size"] == "5,6"


def test_Digraph_pipe(graph):
    if DOT_IS_AVAILABLE:
        with pytest.raises(ValueError):
            graph.pipe(format="12345678")

        assert graph._repr_png_() == graph.pipe(format="png")
    else:
        with pytest.raises(FileNotFoundError):
            graph.pipe(format="png")


@pytest.mark.skipif(not DOT_IS_AVAILABLE, reason="requires dot")
def test_Digraph_repr_png(graph):
    assert graph._repr_png_() == graph.pipe(format="png")


@pytest.mark.skipif(not DOT_IS_AVAILABLE, reason="requires dot")
def test_Digraph_repr(graph):
    from pyhepmc._graphviz import Digraph, Block

    s = repr(graph)
    graph2 = eval(s, {"Digraph": Digraph, "Block": Block})
    assert graph == graph2
    graph2.graph_attr["foo"] = "bar"
    assert graph != graph2


@pytest.mark.skipif(not DOT_IS_AVAILABLE, reason="requires dot")
def test_repr_html(graph, evt):
    assert evt._repr_html_() == graph._repr_html_()


@pytest.mark.skipif(not DOT_IS_AVAILABLE, reason="requires dot")
@pytest.mark.parametrize("ext", view.SUPPORTED_FORMATS)
def test_savefig_1(evt, ext):
    fname = RESULT_DIR / f"test_savefig_1.{ext}"
    view.savefig(evt, fname)

    with io.BytesIO() as f2:
        g = view.to_dot(evt)
        view.savefig(g, f2, format=ext)
        f2.flush()
        f2.seek(0)
        with open(fname, "rb") as f1:
            # Both buffers should be almost equal.
            # They may differ in the creation date,
            # which should be less than 150 bytes
            a1 = np.frombuffer(f1.read(), np.uint8)
            a2 = np.frombuffer(f2.read(), np.uint8)
            size = min(len(a1), len(a2))
            assert size > 0
            assert np.sum(a1[:size] != a2[:size]) < 150


def test_savefig_2(evt):
    with pytest.raises(ValueError):
        view.savefig(evt, "foo")

    with pytest.raises(ValueError):
        view.savefig(evt, "foo.foo")

    with pytest.raises(ValueError):
        with io.BytesIO() as f:
            view.savefig(evt, f)


@pytest.mark.skipif("CI" in os.environ, reason="does not work on CI")
@pytest.mark.skipif(not DOT_IS_AVAILABLE, reason="requires dot")
@pytest.mark.parametrize("ext", ("pdf", "png", "svg"))
def test_savefig_3(evt, ext):
    pytest.importorskip("particle")
    testing = pytest.importorskip("matplotlib.testing")
    compare = pytest.importorskip("matplotlib.testing.compare")
    testing.set_font_settings_for_testing()
    testing.set_reproducibility_for_testing()
    fname = f"test_savefig_3.{ext}"
    expected = REFERENCE_DIR / fname
    actual = RESULT_DIR / fname
    view.savefig(evt, actual)
    assert compare.compare_images(expected, actual, 1e-3) is None


@pytest.mark.skipif(not DOT_IS_AVAILABLE, reason="requires dot")
def test_savefig_4(evt):
    with io.BytesIO() as f:
        g = view.to_dot(evt)
        with pytest.warns(RuntimeWarning):
            view.savefig(g, f, format="png", color_hadron="green")



"""
pyhepmc is a pythonic and Jupyter-friendly Python API for HepMC3.

Differences between HepMC3 C++ and pyhepmc
------------------------------------------

- The pyhepmc API uses properties where the C++ API uses setters/getters
  (where possible).
- Sequences with matching types and lengths are implicitly
  convertible to :class:`FourVector` und :class:`ToolInfo`.
- In addition to the C++ Reader/Writer classes, we offer an easy to use
  :func:`open`. It can read and write any supported HepMC3 format,
  including compressed files (gzip, bzip2, lzma are supported).
- Attributes for :class:`GenRunInfo`, :class:`GenEvent`, :class:`GenParticle`,
  :class:`GenVertex` can be accessed via a dict-like view returned by the
  ``attributes`` property. The view automatically converts between native C++
  types to native Python types.
- The ``Print`` class is missing, but :func:`listing` and :func:`content`
  are present as free functions.
- The member functions ``delta_X`` of :class:`FourVector` are free functions
  with two arguments.
- ``HEPEVT_Wrapper`` and friends are missing, use :meth:`GenEvent.from_hepevt`
  instead.
- ``ReaderGZ`` and ``WriterGZ`` are missing, since :func:`open` offers this
  functionality.
- API marked as deprecated in HepMC3 is not available in Python.
- pyhepmc offers event visualization and renders in Jupyter notebooks if all
  required extra packages are installed, see :func:`pyhepmc.view.to_dot`.

Missing functionality
---------------------

- ``ReaderMT`` will not be implemented. If you want to use multi-threaded IO,
  it is better to use the high-level threading API of Python to achieve this.
  You can read the next GenEvent in a child thread from a file while you are
  processing the current event in the main thread.

"""
from pyhepmc._core import (
    Units,
    FourVector,
    GenEventData,
    GenEvent,
    GenParticle,
    GenVertex,
    GenHeavyIon,
    GenRunInfo,
    GenPdfInfo,
    GenCrossSection,
    HEPRUPAttribute,
    HEPEUPAttribute,
    equal_vertex_sets,
    equal_particle_sets,
    content,
    listing,
    delta_phi,
    delta_eta,
    delta_r2_eta,
    delta_r_eta,
    delta_r2_rap,
    delta_r_rap,
    delta_rap,
)
from pyhepmc.io import open as open  # noqa: F401
from pyhepmc import _attributes
from pyhepmc._setup import Setup
from pyhepmc.view import to_dot
from typing import Any
from importlib.metadata import version

__version__ = version("pyhepmc")

__all__ = (
    "Units",
    "FourVector",
    "GenEventData",
    "GenParticleData",
    "GenVertexData",
    "GenEvent",
    "GenParticle",
    "GenVertex",
    "GenHeavyIon",
    "GenRunInfo",
    "GenPdfInfo",
    "GenCrossSection",
    "HEPRUPAttribute",
    "HEPEUPAttribute",
    "equal_vertex_sets",
    "equal_particle_sets",
    "content",
    "listing",
    "Setup",
    "delta_phi",
    "delta_eta",
    "delta_r2_eta",
    "delta_r_eta",
    "delta_r2_rap",
    "delta_r_rap",
    "delta_rap",
    "open",
)

_attributes.install()

GenEvent._repr_html_ = lambda self: to_dot(self)._repr_html_()


def __getattr__(name: str) -> Any:
    from . import io
    import warnings
    from numpy import VisibleDeprecationWarning

    if name in dir(io):
        warnings.warn(
            f"importing {name} from pyhepmc is deprecated, please import from pyhepmc.io",
            category=VisibleDeprecationWarning,
            stacklevel=2,
        )
        return getattr(io, name)

    raise AttributeError



import pyhepmc
from pyhepmc._core import _Setup_print_errors, _Setup_print_warnings, _Setup_debug_level
import pytest


def test_all():
    assert pyhepmc.Setup.print_errors
    assert pyhepmc.Setup.print_warnings
    assert pyhepmc.Setup.debug_level == 5

    pyhepmc.Setup.print_errors = False
    assert _Setup_print_errors() is False

    pyhepmc.Setup.print_warnings = False
    assert _Setup_print_warnings() is False

    pyhepmc.Setup.debug_level = 10
    assert _Setup_debug_level() == 10

    with pytest.raises(AttributeError):
        pyhepmc.Setup.foo = 1



"""
HepMC3 IO classes.

This module contains various Reader and Writer classes, which have
pythonic interfaces. They act as context managers and close the file
automatically. The Readers can be iterated over and yield events.

The :func:`open` function is even easier to use. It can read any
supported file and will auto-detect the format. It can be used for
reading and writing.
"""
from __future__ import annotations
from ._core import (
    GenEvent,
    ReaderAscii as ReaderAsciiBase,
    ReaderAsciiHepMC2 as ReaderAsciiHepMC2Base,
    ReaderLHEF as ReaderLHEFBase,
    ReaderHEPEVT as ReaderHEPEVTBase,
    WriterAscii,
    WriterAsciiHepMC2,
    WriterHEPEVT,
    UnparsedAttribute,
    pyiostream,
)
from pathlib import PurePath
from typing import Union, Any, Optional, Callable

__all__ = [
    "open",
    "ReaderAscii",
    "ReaderAsciiHepMC2",
    "ReaderLHEF",
    "ReaderHEPEVT",
    "WriterAscii",
    "WriterAsciiHepMC2",
    "WriterHEPEVT",
    "UnparsedAttribute",
]


def _enter(self: Any) -> Any:
    return self


def _exit_close(self: Any, type: Exception, value: str, tb: Any) -> bool:
    self.close()
    return False


def _exit_flush(self: Any, type: Exception, value: str, tb: Any) -> bool:
    self.flush()
    return False


def _read_event_lhef_patch(self: Any, evt: GenEvent) -> bool:
    failed = self._read_event_unpatched(evt)
    if failed and self.failed():  # probably EOF
        return True
    return not failed


class _Iter:
    def __init__(self, reader: Any):
        self.reader = reader

    def __next__(self) -> GenEvent:
        evt = self.reader.read()
        if evt is None:
            raise StopIteration
        return evt

    def __iter__(self) -> "_Iter":
        return self

    next = __next__


class ReaderMixin:
    def read(self) -> Optional[GenEvent]:
        assert hasattr(self, "failed")
        assert hasattr(self, "read_event")
        if self.failed():
            # usually this means EOF was reached previously
            return None
        evt = GenEvent()
        success = self.read_event(evt)
        # workaround for bug in HepMC3, which reports success even if
        # the next section of the file does not contain any event data
        if len(evt.particles) == 0:
            success = False
        return evt if success else None

    def __iter__(self: Any) -> _Iter:
        return _Iter(self)

    __enter__ = _enter
    __exit__ = _exit_close


# add contextmanager interface to IO classes
class ReaderAscii(ReaderAsciiBase, ReaderMixin):  # type:ignore
    pass


class ReaderAsciiHepMC2(ReaderAsciiHepMC2Base, ReaderMixin):  # type:ignore
    pass


class ReaderLHEF(ReaderLHEFBase, ReaderMixin):  # type:ignore
    _read_event_unpatched = ReaderLHEFBase.read_event
    read_event = _read_event_lhef_patch


class ReaderHEPEVT(ReaderHEPEVTBase, ReaderMixin):  # type:ignore
    pass


WriterAscii.__enter__ = _enter
WriterAscii.__exit__ = _exit_close
WriterAscii.write = WriterAscii.write_event

WriterAsciiHepMC2.__enter__ = _enter
WriterAsciiHepMC2.__exit__ = _exit_close
WriterAsciiHepMC2.write = WriterAsciiHepMC2.write_event

WriterHEPEVT.__enter__ = _enter
WriterHEPEVT.__exit__ = _exit_close
WriterHEPEVT.write = WriterHEPEVT.write_event

pyiostream.__enter__ = _enter
pyiostream.__exit__ = _exit_flush


Filename = Union[str, PurePath]


class _WrappedWriter:
    # Wrapper for Writer, to be used by `open`

    def __init__(
        self,
        iostream: Any,
        precision: Optional[int],
        Writer: Any,
    ):
        self._writer: Any = None
        self._init = (iostream, precision, Writer)
        self._event = None

    def _maybe_convert(self, event: Any) -> GenEvent:
        if isinstance(event, GenEvent):
            return event
        if hasattr(event, "to_hepmc3"):
            # reuse GenEvent to not recreate GenRunInfo repeatedly
            self._event = event.to_hepmc3(self._event)
            return self._event
        raise TypeError(
            "event must be an instance of GenEvent or "
            "convertible to it by providing a to_hepmc3() method"
        )

    def write(self, event: Any) -> None:
        evt = self._maybe_convert(event)

        if self._writer is None:
            # first call
            iostream, precision, Writer = self._init
            if Writer is WriterHEPEVT:
                self._writer = Writer(iostream)
            else:
                self._writer = Writer(iostream, evt.run_info)
            if precision is not None and hasattr(self._writer, "precision"):
                self._writer.precision = precision

        self._writer.write_event(evt)
        if self._writer.failed():
            raise IOError("writing GenEvent failed")

    def close(self) -> None:
        if self._writer is not None:
            self._writer.close()

    __enter__ = _enter
    __exit__ = _exit_close


class HepMCFile:
    """
    HepMC file for reading or writing.

    Parameters
    ----------
    fileobj : str or Path or IO object
        Filename to open for reading or writing or file object. When writing to
        existing files, the contents are replaced. When the filename ends with the
        suffixes ".gz", ".bz2", or ".xz", the contents are transparently compressed
        and decompressed.
    mode : str, optional
        Must be either "r" (default) or "w", to indicate whether to open for reading
        or writing.
    precision : int or None, optional
        How many digits of precision to use when writing to a file. Can be used to
        improve the compression rate.
    format : str or None, optional
        Which format to use for reading or writing. If None (default), autodetect
        format when reading (this is fast and thus safe to use), and use the latest
        HepMC3 format when writing. Allowed values (case-insensitive): "HepMC3",
        "HepMC2", "LHEF", "HEPEVT". "LHEF" is not supported for writing.

    Raises
    ------
    IOError if reading or writing fails.
    """

    def __init__(
        self,
        fileobj: Filename,
        mode: str = "r",
        precision: int = None,
        format: str = None,
    ):
        open_file: Optional[Callable[[], Any]] = None
        if hasattr(fileobj, "read") and hasattr(fileobj, "write"):
            if hasattr(fileobj, "buffer"):
                self._file = fileobj.buffer
            else:
                self._file = fileobj
            self._close_file = False
        else:
            fn = str(fileobj)

            if fn.endswith(".gz"):
                import gzip

                open = gzip.open
            elif fn.endswith(".bz2"):
                import bz2

                open = bz2.open  # type:ignore
            elif fn.endswith(".xz"):
                import lzma

                open = lzma.open  # type:ignore
            else:
                from builtins import open  # type:ignore

                mode += "b"

            def open_file() -> Any:
                return open(fn, mode)

            self._close_file = True

        if mode.startswith("r"):
            if open_file:
                self._file = open_file()
            self._ios = pyiostream(self._file)

            Reader: Optional[Any] = None
            if format is None:
                # auto-detect
                if not self._file.seekable():
                    raise ValueError("cannot detect format, file is not seekable")
                header = self._file.read(256)
                assert isinstance(header, bytes)  # for mypy
                self._file.seek(0)
                if b"HepMC::Asciiv3" in header:
                    Reader = ReaderAscii
                elif b"HepMC::IO_GenEvent" in header:
                    Reader = ReaderAsciiHepMC2
                elif b"<LesHouchesEvents" in header:
                    Reader = ReaderLHEF
                else:
                    # this one has no header
                    Reader = ReaderHEPEVT
            else:
                Reader = {
                    "hepmc3": ReaderAscii,
                    "hepmc2": ReaderAsciiHepMC2,
                    "lhef": ReaderLHEF,
                    "hepevt": ReaderHEPEVT,
                }.get(format.lower(), None)
                if Reader is None:
                    raise ValueError(f"format {format!r} not recognized for reading")

            self._reader = Reader(self._ios)
            self._writer = None

        elif mode.startswith("w"):
            if format is None:
                Writer = WriterAscii
            else:
                Writer = {
                    "hepmc3": WriterAscii,
                    "hepmc2": WriterAsciiHepMC2,
                    "hepevt": WriterHEPEVT,
                }.get(format.lower(), None)
                if Writer is None:
                    raise ValueError(f"format {format!r} not recognized for writing")

            if open_file:
                self._file = open_file()
            self._ios = pyiostream(self._file)
            self._reader = None
            self._writer = _WrappedWriter(self._ios, precision, Writer)
        else:
            raise ValueError(f"mode must be 'r' or 'w', got {mode!r}")

    __enter__ = _enter
    __exit__ = _exit_close

    def __iter__(self) -> Any:
        return self._reader.__iter__()

    def flush(self) -> None:
        if not self._writer:
            raise IOError("File opened for reading")
        self._ios.flush()
        self._file.flush()

    def read(self) -> GenEvent:
        if not self._reader:
            raise IOError("File openened for writing")
        return self._reader.read()

    def write(self, event: GenEvent) -> None:
        if not self._writer:
            raise IOError("File openened for reading")
        self._writer.write(event)

    def close(self) -> None:
        if self._reader:
            self._reader.close()  # type:ignore
        if self._writer:
            self._writer.close()
        self._ios.flush()
        if self._close_file:
            self._file.close()


def open(
    fileobj: Filename,
    mode: str = "r",
    precision: int = None,
    format: str = None,
) -> Any:
    """
    Open HepMC files for reading or writing.

    See HepMCFile.
    """
    return HepMCFile(fileobj, mode, precision, format)



import warnings
from numpy import VisibleDeprecationWarning
import typing as _tp


class deprecated:
    def __init__(self, reason: str):
        if not isinstance(reason, str):
            raise ValueError("argument `reason: str` is required")
        self._reason = reason

    def __call__(self, func: _tp.Callable[..., _tp.Any]) -> _tp.Callable[..., _tp.Any]:
        def decorated_func(*args: _tp.Any, **kwargs: _tp.Any) -> _tp.Any:
            warnings.warn(
                f"{func.__name__} is deprecated: {self._reason}",
                category=VisibleDeprecationWarning,
                stacklevel=2,
            )
            return func(*args, **kwargs)

        decorated_func.__name__ = func.__name__
        decorated_func.__doc__ = "deprecated: " + self._reason
        return decorated_func



import os
import pyhepmc as hep
import pyhepmc.io as io
import pytest
from test_basic import make_evt
from pyhepmc._core import stringstream, pyiostream
from io import BytesIO
from pathlib import Path
import numpy as np
import typing
import gzip
from sys import version_info
import subprocess as subp

if version_info >= (3, 9):
    list_type = list
else:
    list_type = typing.List

# this only does something if pyhepmc is compiled in debug mode
hep.Setup.print_warnings = True


@pytest.fixture()
def evt():
    return make_evt()


def test_pystream_1():
    fn = str(Path(__file__).parent / "sibyll21.dat")
    with open(fn, "rb") as f:
        with pyiostream(f, 1000) as s:
            with io.ReaderAscii(s) as r:
                ev1 = r.read()
    with io.ReaderAscii(fn) as r:
        ev2 = r.read()

    assert ev1.particles == ev2.particles
    assert ev1.vertices == ev2.vertices
    assert ev1.run_info == ev2.run_info


def test_pystream_2():
    fn = str(Path(__file__).parent / "sibyll21.dat")
    fn2 = "sibyll21.dat.gz"

    with open(fn, "rb") as f:
        with gzip.open(fn2, "w") as f2:
            for evt in f:
                f2.write(evt)

    ev1 = []
    with gzip.open(fn2) as f:
        with pyiostream(f, 1000) as s:
            with io.ReaderAscii(s) as r:
                for evt in r:
                    ev1.append(evt)

    assert len(ev1) == 1

    ev2 = []
    with io.ReaderAscii(fn) as r:
        for evt in r:
            ev2.append(evt)

    assert ev1 == ev2

    os.unlink(fn2)


def test_pystream_3(evt):
    fn = "test_pystream_3.dat.gz"
    with gzip.open(fn, "w") as f:
        with pyiostream(f, 1000) as s:
            with io.WriterAscii(s) as w:
                w.write(evt)

    with gzip.open(fn) as f:
        with pyiostream(f, 1000) as s:
            with io.ReaderAscii(s) as r:
                evt2 = r.read()

    assert evt == evt2

    os.unlink(fn)


@pytest.mark.parametrize("size", (100, 1))
def test_pystream_4(size):
    s = b"\r\n1\r\n\r\n22\r\n333\r\n"
    io = BytesIO(s)
    pio = pyiostream(io, size)
    assert pio.getline() == b""
    assert pio.getline() == b"1"
    assert pio.getline() == b""
    assert pio.getline() == b"22"
    assert pio.getline() == b"333"
    assert pio.getline() == b""
    assert pio.getline() == b""


def test_pystream_5():
    import sys

    io = sys.stdout.buffer
    with pyiostream(io, 100) as pio:
        pio.write(b"foo")


def test_read_event_write_event(evt):
    oss = stringstream()
    with io.WriterAscii(oss) as f:
        f.write_event(evt)

    evt2 = hep.GenEvent()
    evt3 = hep.GenEvent()
    assert evt != evt2
    with io.ReaderAscii(oss) as f:
        assert f.read_event(evt2)
        assert not f.failed()
        success = f.read_event(evt3)
        assert success  # for EOF success is true and failed() is true
        assert f.failed()

    assert evt.event_number == evt2.event_number
    assert evt.momentum_unit == evt2.momentum_unit
    assert evt.length_unit == evt2.length_unit
    assert evt.particles == evt2.particles
    assert evt.vertices == evt2.vertices
    assert evt == evt2


def test_pythonic_read_write_from_stream(evt):
    oss = stringstream()
    with io.WriterAscii(oss) as f:
        f.write(evt)

    with io.ReaderAscii(oss) as f:
        n = 0
        for evt2 in f:
            assert evt.particles == evt2.particles
            assert evt.vertices == evt2.vertices
            assert evt.run_info == evt2.run_info
            assert evt == evt2
            n += 1
            break
        assert n == 1

        for evt2 in f:
            n += 1
        assert n == 1


def test_failed_read_file():
    with io.ReaderAscii("file_does_not_exist.dat") as f:
        evt = f.read()

    assert evt is None

    n = 0
    with io.ReaderAscii("file_does_not_exist.dat") as f:
        for ev in f:
            n += 1
    assert n == 0


def test_failed_read_file_2():
    fn = str(Path(__file__).parent / "broken.dat")
    with io.ReaderAscii(fn) as f:
        event = f.read()

    assert event is None

    n = 0
    with io.ReaderAscii(fn) as f:
        for ev in f:
            n += 1
    assert n == 0


def test_read_empty_stream():
    oss = stringstream()
    with io.ReaderAscii(oss) as f:
        ev = hep.GenEvent()
        success = f.read_event(ev)
        # reading empty stream is EOF
        assert success
        assert f.failed()

    with io.ReaderAscii(oss) as f:
        ev = f.read()
        # reading empty stream is EOF
        assert ev is None
        assert f.failed()


@pytest.mark.parametrize("format", ("hepmc3", "hepmc2", "hepevt"))
def test_open_1(evt, format):
    with hep.open("test_read_write_file.dat", "w", format=format) as f:
        f.write(evt)

    with hep.open("test_read_write_file.dat", format=format) as f:
        evt2 = f.read()

    if format in ("hepmc2", "hepevt"):
        # ToolInfo not stored in this format, so adding it manually
        evt2.run_info.tools = evt.run_info.tools

    assert evt == evt2

    with hep.open("test_read_write_file.dat") as f:
        evt3 = f.read()

    if format in ("hepmc2", "hepevt"):
        # ToolInfo not stored in this format, so adding it manually
        evt3.run_info.tools = evt.run_info.tools

    assert evt == evt3

    os.unlink("test_read_write_file.dat")


def test_open_2(evt):
    filename = "test_read_write_file.dat"
    with hep.open(filename, "w", precision=3) as f:
        f.write(evt)

    with hep.open(filename) as f:
        for i, evt2 in enumerate(f):
            assert i == 0
            pass

    assert evt != evt2

    with hep.open(filename, "w") as f:
        f.write(evt)

    with hep.open(filename) as f:
        for i, evt3 in enumerate(f):
            assert i == 0
            pass

    assert evt == evt3

    os.unlink(filename)


def test_open_3(evt):
    filename = Path("test_read_write_file.dat")

    with hep.open(filename, "w") as f:
        with pytest.raises(TypeError):
            f.write(None)

        with pytest.raises(TypeError):
            f.write("foo")

    class Foo:
        def __init__(self, evt):
            self.event = evt
            self.run_info = evt.run_info

        def to_hepmc3(self, event=None):
            if event is None:
                event = self.event
            assert event.run_info is self.run_info
            return event

    foo = Foo(evt)

    with hep.open(filename, "w") as f:
        f.write(foo)

    with hep.open(filename) as f:
        evt2 = f.read()

    assert evt == evt2

    filename.unlink()


def test_open_4():
    filename = "test_open_4.dat"

    with hep.open(filename, "w") as f:
        for i in range(3):
            ev = hep.GenEvent()
            ev.event_number = i + 1
            for k in range(ev.event_number):
                ev.add_particle(hep.GenParticle((1, 2, 3, 4), 1, 1))
            f.write(ev)

    with hep.open(filename) as f:
        n = 0
        for ev in f:
            n += 1
            assert ev.event_number == n
            assert len(ev.particles) == n
        assert n == 3

    # iterator is itself iterable
    with hep.open(filename) as f:
        fiter = iter(f)
        n = 0
        for ev in fiter:
            n += 1
        assert n == 3

    os.unlink(filename)


def test_open_5():
    fn = Path(__file__).parent / "pp.lhe"
    n = 0
    with hep.open(fn) as f:
        for ev in f:
            n += 1
    assert n == 1


def test_open_6(evt, capsys):
    import sys

    fn = "test_open_6.dat"

    with hep.open(fn, "w") as f:
        f.write(evt)

    with open(fn, "rb") as f:
        content = f.read().decode()

    with hep.open(sys.stdout, "w") as f:
        f.write(evt)

    c = capsys.readouterr().out
    assert c == content

    os.unlink(fn)


def test_open_7():
    fn = str(Path(__file__).parent / "sibyll21.dat")

    with open(fn, "r") as f:
        with hep.open(f, "r") as f2:
            evt = f2.read()

    assert len(evt.particles) == 23
    assert len(evt.vertices) == 7


def test_open_last_event_issue():
    fn = str(Path(__file__).parent / "last_event_issue.hepmc")

    n = 0
    with io.open(fn) as f:
        while True:
            event = f.read()
            if event is None:
                break
            n += 1
            assert len(event.particles) > 0

    assert n == 1


def test_ReaderAscii_last_event_issue():
    fn = str(Path(__file__).parent / "last_event_issue.hepmc")

    n = 0
    evt = hep.GenEvent()
    with io.ReaderAscii(fn) as f:
        while not f.failed():
            if f.read_event(evt):
                n += 1

    assert n == 1


def test_open_failures():
    fn = Path(__file__).parent / "pp.lhe"
    n = 0
    with hep.open(fn, format="hepmc3") as f:
        # ReaderAscii just skips unreadable parts of file
        for ev in f:
            n += 1
    assert n == 0

    with pytest.raises(ValueError, match="format"):
        with hep.open(fn, format="foo") as f:
            pass

    with pytest.raises(ValueError, match="format"):
        with hep.open("test.dat", "w", format="foo") as f:
            pass

    with pytest.raises(ValueError, match="mode"):
        with hep.open("test.dat", "x") as f:
            pass

    with pytest.raises(ValueError, match="mode"):
        with hep.open("test.dat", "rb") as f:
            pass


@pytest.mark.skipif(
    "CIBW" in os.environ,
    reason=(
        "does not work in cibuildwheel, "
        "although it works in a manually set-up docker"
    ),
)
def test_open_on_readonly_file():
    foo = Path("foo.dat")
    foo.touch(mode=0o000)  # not writeable

    with pytest.raises(IOError):
        with hep.open(foo, "w") as f:
            f.write(hep.GenEvent())

    foo.chmod(mode=0o666)
    foo.unlink()


def test_open_broken():
    fn = Path(__file__).parent / "broken.dat"
    n = 0
    with hep.open(fn) as f:
        for ev in f:
            n += 1
    assert n == 0


@pytest.mark.parametrize(
    "writer", (io.WriterAscii, io.WriterAsciiHepMC2, io.WriterHEPEVT)
)
def test_open_with_writer(evt, writer):  # noqa
    filename = f"test_open_{writer.__name__}.dat"
    with writer(filename) as f:
        f.write(evt)

    with hep.open(filename) as f:
        evt2 = f.read()

    if writer in (io.WriterAsciiHepMC2, io.WriterHEPEVT):
        # ToolInfo not stored in this format, so adding it manually
        evt2.run_info.tools = evt.run_info.tools

    assert evt == evt2

    os.unlink(filename)


def test_open_standalone(evt):  # noqa
    filename = "test_open_standalone.dat"

    f = hep.open(filename, "w")
    f.write(evt)
    f.flush()
    f.close()

    f2 = hep.open(filename)
    evt2 = f2.read()
    f2.close()

    assert evt2 == evt

    f3 = hep.io.HepMCFile(filename)
    for evt3 in f3:
        pass
    f3.close()

    assert evt3 == evt

    os.unlink(filename)


def test_deprecated_import():
    with pytest.warns(np.VisibleDeprecationWarning):
        from pyhepmc import ReaderAscii  # noqa F401


def test_attributes():
    filename = "test_attributes.dat"

    evt = hep.GenEvent()  # noqa
    p = hep.GenParticle((1, 2, 3, 4), 5, 6)
    evt.add_particle(p)
    evt.attributes = {
        "1": True,
        "2": 2,
        "3": 3.3,
        "4": hep.GenPdfInfo(),
        "5": hep.GenHeavyIon(),
        "6": hep.GenCrossSection(),
        "7": p,
        "8": [1, 2],
        "9": [1.3, 2.4],
        # cannot test this yet
        # "8": hep.HEPRUPAttribute(),
        # "9": hep.HEPEUPAttribute(),
    }

    with io.open(filename, "w") as f:
        f.write(evt)

    with io.open(filename) as f:
        evt2 = f.read()

    with pytest.raises(TypeError):
        evt2.attributes["1"].astype(hep.GenPdfInfo)

    with pytest.raises(TypeError, match="untyped list"):
        evt2.attributes["8"].astype(list)

    for k, v in evt.attributes.items():
        if k == "8":
            t = list_type[int]
        elif k == "9":
            t = list_type[float]
        else:
            t = type(v)

        v2 = evt2.attributes[k].astype(t)

        assert v == v2
        # UnparsedAttribute is replaced with parsed Attribute
        v3 = evt2.attributes[k]
        assert v == v3

    os.unlink(filename)


@pytest.mark.parametrize("format", ["hepmc3", "hepmc2", "hepevt"])
def test_roundtrip(evt, format):
    fn = "test_roundtrip.dat"

    if format != "hepmc3":
        # only HepMC3 support non-empty GenRunInfo
        evt.run_info = hep.GenRunInfo()
    # HepMC3 adds "0" to weight_names upon reading event if weight_names is empty
    evt.run_info.weight_names = ["0"]

    with io.open(fn, "w", format=format) as f:
        f.write(evt)

    with io.open(fn, "r", format=format) as f:
        evt2 = f.read()

    os.unlink(fn)

    assert evt == evt2


@pytest.mark.parametrize("zip", ["gz", "bz2", "xz"])
def test_zip(evt, zip):
    fn = f"test_zip.{zip}"
    # HepMC3 adds "0" to weight_names upon reading event if weight_names is empty
    evt.run_info.weight_names = ["0"]

    with io.open(fn, "w") as f:
        f.write(evt)

    try:
        out = subp.check_output(["file", fn], text=True)
        assert "compressed" in out.lower()
        assert {"gz": "gzip", "bz2": "bzip2", "xz": "xz"}[zip] in out.lower()
    except FileNotFoundError:
        pass

    with io.open(fn, "r") as f:
        evt2 = f.read()

    os.unlink(fn)

    assert evt == evt2



# DO NOT EDIT: created by generate_docs.py
# flake8: noqa
autodoc = {
    "AssociatedParticle": """Attribute class allowing eg. a GenParticle to refer to another GenParticle.""",
    "AssociatedParticle.__init__": """Default constructor.

Constructor initializing attribute value.""",
    "AssociatedParticle.associated": """get a pointer to the associated particle.

set the value associated to this Attribute.""",
    "AssociatedParticle.associatedId": """get id of the associated particle.""",
    "AssociatedParticle.event": """return the GenEvent to which this Attribute belongs, if at all.""",
    "AssociatedParticle.from_string": """Implementation of Attribute::from_string.""",
    "AssociatedParticle.init": """Optionally initialize the attribute after from_string.""",
    "AssociatedParticle.is_parsed": """Check if this attribute is parsed.""",
    "AssociatedParticle.particle": """return the GenParticle to which this Attribute belongs, if at all.""",
    "AssociatedParticle.to_string": """Implementation of Attribute::to_string.""",
    "AssociatedParticle.unparsed_string": """Get unparsed string.""",
    "AssociatedParticle.value": """get the value associated to this Attribute.

set the value associated to this Attribute.""",
    "AssociatedParticle.vertex": """return the GenVertex to which this Attribute belongs, if at all.""",
    "FourVector": """Generic 4-vector.

Interpretation of its content depends on accessors used: it's much simpler to do this than to distinguish between space and momentum vectors via the type system (especially given the need for backward compatibility with HepMC2). Be sensible and don't call energy functions on spatial vectors! To avoid duplication, most definitions are only implemented on the spatial function names, with the energy-momentum functions as aliases.

This is not intended to be a fully featured 4-vector, but does contain the majority of common non-boosting functionality, as well as a few support operations on 4-vectors.

The implementations in this class are fully inlined.""",
    "FourVector.__init__": """Default constructor.

Accesss all FourVector fields.

Copy constructor.""",
    "FourVector.abs_eta": """Absolute pseudorapidity.""",
    "FourVector.abs_rap": """Absolute rapidity.""",
    "FourVector.e": """Energy component of momentum.""",
    "FourVector.eta": """Pseudorapidity.""",
    "FourVector.interval": """Spacetime invariant interval s^2 = t^2 - x^2 - y^2 - z^2.""",
    "FourVector.is_zero": """Check if the length of this vertex is zero.""",
    "FourVector.length": """Magnitude of spatial (x, y, z) 3-vector.""",
    "FourVector.length2": """Squared magnitude of (x, y, z) 3-vector.""",
    "FourVector.m": """Invariant mass. Returns -sqrt(-m) if e^2 - P^2 is negative.""",
    "FourVector.m2": """Squared invariant mass m^2 = E^2 - px^2 - py^2 - pz^2.""",
    "FourVector.p3mod": """Magnitude of p3 = (px, py, pz) vector.""",
    "FourVector.p3mod2": """Squared magnitude of p3 = (px, py, pz) vector.""",
    "FourVector.perp": """Magnitude of (x, y) vector.""",
    "FourVector.perp2": """Squared magnitude of (x, y) vector.""",
    "FourVector.phi": """Azimuthal angle.""",
    "FourVector.pt": """Transverse momentum.""",
    "FourVector.pt2": """Squared transverse momentum px^2 + py^2.""",
    "FourVector.px": """x-component of momentum.""",
    "FourVector.py": """y-component of momentum.""",
    "FourVector.pz": """z-component of momentum.""",
    "FourVector.rap": """Rapidity.""",
    "FourVector.rho": """Magnitude of spatial (x, y, z) 3-vector, for HepMC2 compatibility.""",
    "FourVector.t": """Time component of position/displacement.""",
    "FourVector.theta": """Polar angle w.r.t. z direction.""",
    "FourVector.x": """x-component of position/displacement.""",
    "FourVector.y": """y-component of position/displacement.""",
    "FourVector.z": """z-component of position/displacement.""",
    "GenCrossSection": """Stores additional information about cross-section.

This is an example of event attribute used to store cross-section information.

This class is meant to be used to pass, on an event by event basis, the current best guess of the total cross section. It is expected that the final cross section will be stored elsewhere.

double cross_section; // cross section in pb.
double cross_section_error; // error associated with this cross section.
long accepted_events; ///< The number of events generated so far.
long attempted_events; ///< The number of events attempted so far.

In addition, several cross sections and related info can be included in case of runs with mulltiple weights.

The units of cross_section and cross_section_error are expected to be pb.""",
    "GenCrossSection.accepted_events": """Access the number of accepted events.""",
    "GenCrossSection.attempted_events": """Access the number of attempted events.""",
    "GenCrossSection.event": """return the GenEvent to which this Attribute belongs, if at all.""",
    "GenCrossSection.from_string": """Implementation of Attribute::from_string.""",
    "GenCrossSection.init": """Optionally initialize the attribute after from_string.""",
    "GenCrossSection.is_parsed": """Check if this attribute is parsed.""",
    "GenCrossSection.is_valid": """Verify that the instance contains non-zero information.""",
    "GenCrossSection.particle": """return the GenParticle to which this Attribute belongs, if at all.""",
    "GenCrossSection.set_cross_section": """Set all fields.""",
    "GenCrossSection.to_string": """Implementation of Attribute::to_string.""",
    "GenCrossSection.unparsed_string": """Get unparsed string.""",
    "GenCrossSection.vertex": """return the GenVertex to which this Attribute belongs, if at all.""",
    "GenCrossSection.xsec": """Access the cross section corresponding to the weight named wName.

Access the cross section corresponding to the weight with index indx.

Access the cross section corresponding to the weight named wName.

Access the cross section corresponding to the weight with index indx.""",
    "GenCrossSection.xsec_err": """Access the cross section error corresponding to the weight named wName.

Access the cross section error corresponding to the weight with index indx.

Access the cross section error corresponding to the weight named wName.

Access the cross section error corresponding to the weight with index indx.""",
    "GenEvent": """Stores event-related information.

Manages event-related information. Contains lists of GenParticle and GenVertex objects.""",
    "GenEvent.__init__": """Event constructor without a run.

Constructor with associated run.

Copy constructor.""",
    "GenEvent.add_attributes": """Add multiple attributes to event.

This will overwrite existing attributes if attributes with the same names are present.""",
    "GenEvent.add_beam_particle": """Add particle to root vertex.""",
    "GenEvent.add_particle": """Add particle.""",
    "GenEvent.add_tree": """Add whole tree in topological order.

This function will find the beam particles (particles that have no production vertices or their production vertices have no particles) and will add the whole decay tree starting from these particles.

Any particles on this list that do not belong to the tree will be ignored.""",
    "GenEvent.add_vertex": """Add vertex.""",
    "GenEvent.attribute": """Get attribute of type T.""",
    "GenEvent.attribute_as_string": """Get attribute of any type as string.""",
    "GenEvent.attribute_names": """Get list of attribute names.""",
    "GenEvent.attributes": """Get a copy of the list of attributes.

To avoid thread issues, this is returns a copy. Better solution may be needed.""",
    "GenEvent.beams": """Vector of beam particles.""",
    "GenEvent.boost": """Boost event using x,y,z components of v as velocities.""",
    "GenEvent.clear": """Remove contents of this event.""",
    "GenEvent.cross_section": """Access cross-section information.""",
    "GenEvent.event_number": """Access event number.""",
    "GenEvent.event_pos": """Vertex representing the overall event position.""",
    "GenEvent.heavy_ion": """Access heavy ion generator additional information.""",
    "GenEvent.length_unit": """Get length unit.""",
    "GenEvent.momentum_unit": """Get momentum unit.""",
    "GenEvent.particles": """Access list of particles.""",
    "GenEvent.particles_empty": """Particles empty, HepMC2 compatiility.""",
    "GenEvent.particles_size": """Particles size, HepMC2 compatiility.""",
    "GenEvent.pdf_info": """Access PDF information.""",
    "GenEvent.read_data": """Fill GenEvent based on GenEventData.""",
    "GenEvent.reflect": """Change sign of axis.""",
    "GenEvent.remove_attribute": """Remove attribute.""",
    "GenEvent.remove_particle": """Remove particle from the event.

This function will remove whole sub-tree starting from this particle if it is the only incoming particle of this vertex. It will also production vertex of this particle if this vertex has no more outgoing particles.""",
    "GenEvent.remove_particles": """Remove a set of particles.

This function follows rules of GenEvent::remove_particle to remove a list of particles from the event.""",
    "GenEvent.remove_vertex": """Remove vertex from the event.

This will remove all sub-trees of all outgoing particles of this vertex.""",
    "GenEvent.reserve": """Reserve memory for particles and vertices.

Helps optimize event creation when size of the event is known beforehand.""",
    "GenEvent.rotate": """Rotate event using x,y,z components of v as rotation angles.""",
    "GenEvent.run_info": """Access a pointer to the the GenRunInfo object.

Access the GenRunInfo object by smart pointer.""",
    "GenEvent.set_units": """Change event units Converts event from current units to new ones.""",
    "GenEvent.shift_position_by": """Shift position of all vertices in the event by delta.""",
    "GenEvent.shift_position_to": """Shift position of all vertices in the event to op.""",
    "GenEvent.vertices": """Access list of vertices.""",
    "GenEvent.vertices_empty": """Vertices empty, HepMC2 compatiility.""",
    "GenEvent.vertices_size": """Vertices size, HepMC2 compatiility.""",
    "GenEvent.weight": """Access event weight accessed by index (or the canonical/first one if there is no argument) It's the user's responsibility to ensure that the given index exists!.

Access event weight accessed by weight name Requires there to be an attached GenRunInfo, otherwise will throw an exception

It's the user's responsibility to ensure that the given name exists!.

Access event weight accessed by weight name Requires there to be an attached GenRunInfo, otherwise will throw an exception

It's the user's responsibility to ensure that the given name exists!.""",
    "GenEvent.weight_names": """Get event weight names, if there are some Requires there to be an attached GenRunInfo with registered weight names, otherwise will throw an exception.""",
    "GenEvent.weights": """Access event weight values as a vector.

Access event weights as a vector.""",
    "GenEvent.write_data": """Fill GenEventData object.""",
    "GenEventData": """Stores serializable event information.""",
    "GenEventData.attribute_id": """Attribute owner id.""",
    "GenEventData.attribute_name": """Attribute name.""",
    "GenEventData.attribute_string": """Attribute serialized as string.""",
    "GenEventData.event_number": """Event number.""",
    "GenEventData.event_pos": """Event position.""",
    "GenEventData.length_unit": """Length unit.""",
    "GenEventData.links1": """First id of the vertex links.

If this id is positive - it is the incoming particle id of a vertex which id is written in GenEventData::links2.

If this id is negative - it's the id of a vertex which outgoing particle id is written in GenEventData::links2.

The links1[i] points to links2[i]. In case links1[i] is particle, links2[i] is end vertex. In case links2[i] is vertex, links2[i] is outgoing particle. An example of usage is given in documentation.""",
    "GenEventData.links2": """Second id of the vertex links.""",
    "GenEventData.momentum_unit": """Momentum unit.""",
    "GenEventData.particles": """Particles.""",
    "GenEventData.vertices": """Vertices.""",
    "GenEventData.weights": """Weights.""",
    "GenHeavyIon": """Stores additional information about Heavy Ion generator.

This is an example of event attribute used to store Heavy Ion information.""",
    "GenHeavyIon.N_Nwounded_collisions": """Collisions with a diffractively excited target nucleon.

The number of single diffractive nucleon-nucleon collisions where the target nucleon is excited. A negative value means that the information is not available.""",
    "GenHeavyIon.Ncoll": """the number of inelastic nucleon-nucleon collisions.

Note that a one participating nucleon can be involved in many inelastic collisions, and that inelastic also includes diffractive excitation. A negative value means that the information is not available.""",
    "GenHeavyIon.Ncoll_hard": """the number of hard nucleon-nucleon collisions.

Model-dependent. Usually the number of nucleon-nucleon collisions containing a special signal process. A negative value means that the information is not available.""",
    "GenHeavyIon.Npart_proj": """the number of participating nucleons in the projectile.

The number of nucleons in the projectile participating in an inelastic collision (see Ncoll). A negative value means that the information is not available.""",
    "GenHeavyIon.Npart_targ": """the number of participating nucleons in the target.

The number of nucleons in the target participating in an inelastic collision (see Ncoll). A negative value means that the information is not available.""",
    "GenHeavyIon.Nspec_proj_n": """The number of spectator neutrons in the projectile.

ie. those that thave not participated in any inelastic nucleon-nucleon collision. A negative value indicatess that the information is not available.""",
    "GenHeavyIon.Nspec_proj_p": """The number of spectator protons in the projectile.

ie. those that thave not participated in any inelastic nucleon-nucleon collision. A negative value indicatess that the information is not available.""",
    "GenHeavyIon.Nspec_targ_n": """The number of spectator neutrons in the target.

ie. those that thave not participated in any inelastic nucleon-nucleon collision. A negative value indicatess that the information is not available.""",
    "GenHeavyIon.Nspec_targ_p": """The number of spectator protons in the target.

ie. those that thave not participated in any inelastic nucleon-nucleon collision. A negative value indicatess that the information is not available.""",
    "GenHeavyIon.Nwounded_N_collisions": """Collisions with a diffractively excited projectile nucleon.

The number of single diffractive nucleon-nucleon collisions where the projectile nucleon is excited. A negative value means that the information is not available.""",
    "GenHeavyIon.Nwounded_Nwounded_collisions": """Non-diffractive or doubly diffractive collisions.

The number of nucleon-nucleon collisions where both projectile and target nucleons are wounded. A negative value means that the information is not available.""",
    "GenHeavyIon.__init__": """Empty default constructor.""",
    "GenHeavyIon.centrality": """The centrality.

The generated centrality in percentiles, where 0 is the maximally central and 100 is the minimally central. A negative value means that the information is not available.""",
    "GenHeavyIon.eccentricities": """Eccentricities.

Calculated to different orders. The key of the map specifies the order, and the value gives the corresponding eccentricity.""",
    "GenHeavyIon.eccentricity": """The eccentricity.

HEPMC3_DEPRECATED("Use eccentricities insted.").""",
    "GenHeavyIon.event": """return the GenEvent to which this Attribute belongs, if at all.""",
    "GenHeavyIon.event_plane_angle": """The event plane angle.

The angle wrt. the x-axix of the impact parameter vector (pointing frm the target to the projectile). A positive number between 0 and two pi. A negative value means that the information is not available.""",
    "GenHeavyIon.forceoldformat": """force writing in old format for compatibility purposes.

HEPMC3_DEPRECATED("This should really not be needed");.""",
    "GenHeavyIon.from_string": """Implementation of Attribute::from_string.""",
    "GenHeavyIon.impact_parameter": """The impact parameter.

The impact parameter given in units of femtometer. A negative value means that the information is not available.""",
    "GenHeavyIon.init": """Optionally initialize the attribute after from_string.""",
    "GenHeavyIon.is_parsed": """Check if this attribute is parsed.""",
    "GenHeavyIon.is_valid": """Verify that the instance contains non-zero information.

HEPMC3_DEPRECATED("Each filed now have default values meaning
that they have not been set").""",
    "GenHeavyIon.participant_plane_angles": """Participant plane angles.

calculated to different orders. The key of the map specifies the order, and the value gives to the angle wrt. the event plane.""",
    "GenHeavyIon.particle": """return the GenParticle to which this Attribute belongs, if at all.""",
    "GenHeavyIon.set": """Set all fields.

HEPMC3_DEPRECATED("Set individual fields directly instead.").

Set all fields.""",
    "GenHeavyIon.sigma_inel_NN": """The assumed inelastic nucleon-nucleon cross section.

in units of millibarn. As used in a Glauber calculation to simulate the distribution in Ncoll. A negative value means that the information is not available.""",
    "GenHeavyIon.spectator_neutrons": """Total number of spectator neutrons.

HEPMC3_DEPRECATED("Use Nspec_proj_n and Nspec_targ_n instead.").""",
    "GenHeavyIon.spectator_protons": """Total number of spectator protons.

HEPMC3_DEPRECATED("Use Nspec_proj_p and Nspec_targ_p instead.").""",
    "GenHeavyIon.to_string": """Implementation of Attribute::to_string.""",
    "GenHeavyIon.unparsed_string": """Get unparsed string.""",
    "GenHeavyIon.user_cent_estimate": """A user defined centrality estimator.

This variable may contain anything a generator feels is reasonable for estimating centrality. The value should be non-negative, and a low value corresponds to a low centrality. A negative value indicatess that the information is not available.""",
    "GenHeavyIon.vertex": """return the GenVertex to which this Attribute belongs, if at all.""",
    "GenParticle": """Stores particle-related information.""",
    "GenParticle.__init__": """Default constructor.

Constructor based on particle data.""",
    "GenParticle.abs_pid": """Get absolute value of PDG ID.""",
    "GenParticle.add_attribute": """Add an attribute to this particle.

This will overwrite existing attribute if an attribute with the same name is present. The attribute will be stored in the parent_event(). false if there is no parent_event();.""",
    "GenParticle.attribute": """Get attribute of type T.""",
    "GenParticle.attribute_as_string": """Get attribute of any type as string.""",
    "GenParticle.attribute_names": """Get list of names of attributes assigned to this particle.""",
    "GenParticle.children": """Convenience access to immediate outgoing particles via end vertex.

Less efficient than via the vertex since return must be by value (in case there is no vertex).""",
    "GenParticle.data": """Get particle data.""",
    "GenParticle.end_vertex": """Access end vertex.""",
    "GenParticle.generated_mass": """Access generated mass.

This function will return mass as set by a generator/tool. If not set, it will return momentum().m().

Access generated mass.""",
    "GenParticle.id": """Get particle id.""",
    "GenParticle.in_event": """Check if this particle belongs to an event.""",
    "GenParticle.is_generated_mass_set": """Check if generated mass is set.""",
    "GenParticle.momentum": """Access momentum.""",
    "GenParticle.parent_event": """Access parent event.""",
    "GenParticle.parents": """Convenience access to immediate incoming particles via production vertex.

Less efficient than via the vertex since return must be by value (in case there is no vertex).""",
    "GenParticle.pid": """Access PDG ID.""",
    "GenParticle.production_vertex": """Access production vertex.""",
    "GenParticle.remove_attribute": """Remove attribute.""",
    "GenParticle.status": """Access status code.""",
    "GenParticle.unset_generated_mass": """Declare that generated mass is not set.""",
    "GenParticleData": """Stores serializable particle information.""",
    "GenParticleData.is_mass_set": """Check if generated mass is set.""",
    "GenParticleData.mass": """Generated mass (if set).""",
    "GenParticleData.momentum": """Momentum.""",
    "GenParticleData.pid": """PDG ID.""",
    "GenParticleData.status": """Status.""",
    "GenPdfInfo": """Stores additional information about PDFs.

This is an example of event attribute used to store PDF-related information.

Input parton flavour codes id1 & id2 are expected to obey the PDG code conventions, especially g = 21.

The contents of pdf1 and pdf2 are expected to be x*f(x). The LHAPDF set ids are the entries in the first column of http:///projects.hepforge.org/lhapdf/PDFsets.index.""",
    "GenPdfInfo.event": """return the GenEvent to which this Attribute belongs, if at all.""",
    "GenPdfInfo.from_string": """Implementation of Attribute::from_string.""",
    "GenPdfInfo.init": """Optionally initialize the attribute after from_string.""",
    "GenPdfInfo.is_parsed": """Check if this attribute is parsed.""",
    "GenPdfInfo.is_valid": """Verify that the instance contains non-zero information.""",
    "GenPdfInfo.particle": """return the GenParticle to which this Attribute belongs, if at all.""",
    "GenPdfInfo.parton_id": """Parton PDG ID.""",
    "GenPdfInfo.pdf_id": """LHAPDF ID code.""",
    "GenPdfInfo.scale": """Factorisation scale (in GEV).""",
    "GenPdfInfo.set": """Set all fields.""",
    "GenPdfInfo.to_string": """Implementation of Attribute::to_string.""",
    "GenPdfInfo.unparsed_string": """Get unparsed string.""",
    "GenPdfInfo.vertex": """return the GenVertex to which this Attribute belongs, if at all.""",
    "GenPdfInfo.x": """Parton momentum fraction.""",
    "GenPdfInfo.xf": """PDF value.""",
    "GenRunInfo": """Stores run-related information.

Manages run-related information. Contains run-wide attributes.""",
    "GenRunInfo.ToolInfo": """Interrnal struct for keeping track of tools.""",
    "GenRunInfo.ToolInfo.description": """Other information about how the tool was used in the run.""",
    "GenRunInfo.ToolInfo.name": """The name of the tool.""",
    "GenRunInfo.ToolInfo.version": """The version of the tool.""",
    "GenRunInfo.__init__": """Default constructor.

Copy constructor.""",
    "GenRunInfo.add_attribute": """add an attribute This will overwrite existing attribute if an attribute with the same name is present.""",
    "GenRunInfo.attribute": """Get attribute of type T.""",
    "GenRunInfo.attribute_as_string": """Get attribute of any type as string.""",
    "GenRunInfo.attribute_names": """Get list of attribute names.""",
    "GenRunInfo.attributes": """Get a copy of the list of attributes.

To avoid thread issues, this is returns a copy. Better solution may be needed.""",
    "GenRunInfo.has_weight": """Check if a weight name is present.""",
    "GenRunInfo.read_data": """Fill GenRunInfo based on GenRunInfoData.""",
    "GenRunInfo.remove_attribute": """Remove attribute.""",
    "GenRunInfo.tools": """The vector of tools used to produce this run.""",
    "GenRunInfo.weight_index": """Return the index corresponding to a weight name.

-1 if name was not found.""",
    "GenRunInfo.weight_indices": """Returns a copy of indices map.""",
    "GenRunInfo.weight_names": """Access the vector of weight names.

Access the names of the weights in this run.

For consistency, the length of the vector should be the same as the number of weights in the events in the run.""",
    "GenRunInfo.write_data": """Fill GenRunInfoData object.""",
    "GenRunInfoData": """Stores serializable run information.""",
    "GenRunInfoData.attribute_name": """Attribute name.""",
    "GenRunInfoData.attribute_string": """Attribute serialized as string.""",
    "GenRunInfoData.tool_description": """Tool descriptions.""",
    "GenRunInfoData.tool_name": """Tool names.""",
    "GenRunInfoData.tool_version": """Tool versions.""",
    "GenRunInfoData.weight_names": """Weight names.""",
    "GenVertex": """Stores vertex-related information.""",
    "GenVertex.__init__": """Default constructor.

Constructor based on vertex data.""",
    "GenVertex.add_attribute": """Add event attribute to this vertex.

This will overwrite existing attribute if an attribute with the same name is present. The attribute will be stored in the parent_event(). false if there is no parent_event();.""",
    "GenVertex.add_particle_in": """Add incoming particle.""",
    "GenVertex.add_particle_out": """Add outgoing particle.""",
    "GenVertex.attribute": """Get attribute of type T.""",
    "GenVertex.attribute_as_string": """Get attribute of any type as string.""",
    "GenVertex.attribute_names": """Get list of names of attributes assigned to this particle.""",
    "GenVertex.data": """Get vertex data.""",
    "GenVertex.has_set_position": """Check if position of this vertex is set.""",
    "GenVertex.id": """Access the vertex unique identifier.

This is not the same as id() in HepMC v2, which is now status().

set the vertex identifier.""",
    "GenVertex.in_event": """Check if this vertex belongs to an event.""",
    "GenVertex.parent_event": """Access parent event.""",
    "GenVertex.particles_in": """Access list of incoming particles.""",
    "GenVertex.particles_in_size": """Number of incoming particles, HepMC2 compatiility.""",
    "GenVertex.particles_out": """Access list of outgoing particles.""",
    "GenVertex.particles_out_size": """Number of outgoing particles, HepMC2 compatiility.""",
    "GenVertex.position": """Access vertex position.

Returns the position of this vertex. If a position is not set on this vertex, the production vertices of ancestors are searched to find the inherited position. FourVector(0,0,0,0) is returned if no position information is found.

Access vertex position.""",
    "GenVertex.remove_attribute": """Remove attribute.""",
    "GenVertex.remove_particle_in": """Remove incoming particle.""",
    "GenVertex.remove_particle_out": """Remove outgoing particle.""",
    "GenVertex.status": """Access vertex status code.""",
    "GenVertexData": """Stores serializable vertex information.""",
    "GenVertexData.is_zero": """Check if this struct fields are zero.""",
    "GenVertexData.position": """Position in time-space.""",
    "GenVertexData.status": """Vertex status.""",
    "Print": """Provides different printing formats.""",
    "Print.content": """Print content of all GenEvent containers.""",
    "Print.listing": """Print event in listing (HepMC2) format.""",
    "ReaderAscii": """GenEvent I/O parsing for structured text files.""",
    "ReaderAscii.__init__": """Constructor.

The ctor to read from stream.

The ctor to read from stream. Useful for temp. streams.""",
    "ReaderAscii.close": """Close file stream.""",
    "ReaderAscii.failed": """Return status of the stream.""",
    "ReaderAscii.options": """Access options.""",
    "ReaderAscii.read_event": """Load event from file.

Parameters
----------
evt:
    Event to be filled.""",
    "ReaderAscii.run_info": """Access the global GenRunInfo object.""",
    "ReaderAscii.skip": """skip events.""",
    "ReaderAsciiHepMC2": """Parser for HepMC2 I/O files.""",
    "ReaderAsciiHepMC2.__init__": """Default constructor.

The ctor to read from stream.

The ctor to read from temp stream.""",
    "ReaderAsciiHepMC2.close": """Close file stream.""",
    "ReaderAsciiHepMC2.failed": """Return status of the stream.""",
    "ReaderAsciiHepMC2.options": """Access options.""",
    "ReaderAsciiHepMC2.read_event": """Implementation of Reader::read_event.""",
    "ReaderAsciiHepMC2.run_info": """Access the global GenRunInfo object.""",
    "ReaderAsciiHepMC2.skip": """skip events.""",
    "ReaderGZ": """GenEvent I/O parsing for compressed files.""",
    "ReaderGZ.__init__": """Constructor.

The ctor to read from stdin.""",
    "ReaderGZ.close": """Close file stream.""",
    "ReaderGZ.failed": """Return status of the stream.""",
    "ReaderGZ.options": """Access options.""",
    "ReaderGZ.read_event": """Load event from file.

Parameters
----------
evt:
    Event to be filled.""",
    "ReaderGZ.run_info": """Access the global GenRunInfo object.""",
    "ReaderGZ.skip": """skip events.""",
    "ReaderHEPEVT": """GenEvent I/O parsing and serialization for HEPEVT files.""",
    "ReaderHEPEVT.__init__": """Default constructor.

The ctor to read from stream.

The ctor to read from temp stream.""",
    "ReaderHEPEVT.close": """Close file stream.""",
    "ReaderHEPEVT.failed": """Get stream error state.""",
    "ReaderHEPEVT.hepevtbuffer": """Pointer to HEPEVT Fortran common block/C struct.""",
    "ReaderHEPEVT.options": """Access options.""",
    "ReaderHEPEVT.read_event": """Read event from file.""",
    "ReaderHEPEVT.read_hepevt_event_header": """Find and read event header line from file.""",
    "ReaderHEPEVT.read_hepevt_particle": """read particle from file.

Parameters
----------
i:
    Particle id.""",
    "ReaderHEPEVT.run_info": """Access the global GenRunInfo object.""",
    "ReaderHEPEVT.skip": """skip events.""",
    "ReaderLHEF": """GenEvent I/O parsing and serialization for LHEF files.""",
    "ReaderLHEF.__init__": """The ctor to read from stream.

Constructor.

The ctor to read from temp stream.""",
    "ReaderLHEF.close": """Close.""",
    "ReaderLHEF.failed": """State.""",
    "ReaderLHEF.options": """Access options.""",
    "ReaderLHEF.read_event": """Reading event.""",
    "ReaderLHEF.run_info": """Access the global GenRunInfo object.""",
    "ReaderLHEF.skip": """skip events.""",
    "ReaderMT": """Multithreader GenEvent I/O parsing.""",
    "ReaderMT.close": """Close file and/or stream.""",
    "ReaderMT.failed": """Get file and/or stream error state.""",
    "ReaderMT.options": """Access options.""",
    "ReaderMT.read_event": """Fill next event from input into evt.""",
    "ReaderMT.run_info": """Access the global GenRunInfo object.""",
    "ReaderMT.skip": """skip or fast forward reading of some events.""",
    "Setup": """Configuration for HepMC.

Contains macro definitions for printing debug output, feature deprecation, etc. Static class - configuration is shared among all HepMC events and program threads.""",
    "Setup.DOUBLE_EPSILON": """Default threshold for comparing double variables.""",
    "Setup.debug_level": """Access debug level.""",
    "Setup.print_errors": """Access error messages printing flag.""",
    "Setup.print_warnings": """Access warning messages printing flag.""",
    "Units": """Stores units-related enums and conversion functions.

Manages units used by HepMC::GenEvent.""",
    "Units.LengthUnit": """Position units.""",
    "Units.MomentumUnit": """Momentum units.""",
    "Units.convert": """Convert FourVector to different momentum unit.

Convert FourVector to different length unit.""",
    "Units.length_unit": """Get length unit based on its name.""",
    "Units.momentum_unit": """Get momentum unit based on its name.""",
    "Units.name": """Access name of momentum unit.

Access name of length unit.""",
    "WriterAscii": """GenEvent I/O serialization for structured text files.""",
    "WriterAscii.__init__": """Constructor.

If file already exists, it will be cleared before writing.

Constructor from ostream.

Constructor from temp ostream.""",
    "WriterAscii.close": """Close file stream.""",
    "WriterAscii.failed": """Return status of the stream.""",
    "WriterAscii.options": """Access options.""",
    "WriterAscii.precision": """Return output precision.

Access output precision.

So far available range is [2,24]. Default is 16.""",
    "WriterAscii.run_info": """Access the global GenRunInfo object.""",
    "WriterAscii.write_event": """Write event to file.

Parameters
----------
evt:
    Event to be serialized.""",
    "WriterAscii.write_run_info": """Write the GenRunInfo object to file.""",
    "WriterAsciiHepMC2": """GenEvent I/O serialization for structured text files.""",
    "WriterAsciiHepMC2.__init__": """Constructor.

If file already exists, it will be cleared before writing.

Constructor from ostream.

Constructor from temp ostream.""",
    "WriterAsciiHepMC2.close": """Close file stream.""",
    "WriterAsciiHepMC2.failed": """Return status of the stream.""",
    "WriterAsciiHepMC2.options": """Access options.""",
    "WriterAsciiHepMC2.precision": """Return output precision.

Access output precision.

Available range is [2,24]. Default is 16.""",
    "WriterAsciiHepMC2.run_info": """Access the global GenRunInfo object.""",
    "WriterAsciiHepMC2.write_event": """Write event to file.

Parameters
----------
evt:
    Event to be serialized.""",
    "WriterAsciiHepMC2.write_run_info": """Write the GenRunInfo object to file.""",
    "WriterGZ": """GenEvent I/O serialization for compressed files.""",
    "WriterGZ.__init__": """Constructor.

If file already exists, it will be cleared before writing.

Constructor from ostream.""",
    "WriterGZ.close": """Close file stream.""",
    "WriterGZ.failed": """Return status of the stream.""",
    "WriterGZ.options": """Access options.""",
    "WriterGZ.run_info": """Access the global GenRunInfo object.""",
    "WriterGZ.write_event": """Write event to file.

Parameters
----------
evt:
    Event to be serialized.""",
    "WriterHEPEVT": """GenEvent I/O serialization for HEPEVT files.""",
    "WriterHEPEVT.__init__": """Default constructor.

If file exists, it will be overwritten.

Constructor from ostream.

Constructor from temp ostream.""",
    "WriterHEPEVT.close": """Close file stream.""",
    "WriterHEPEVT.failed": """Get stream error state flag.""",
    "WriterHEPEVT.options": """Access options.""",
    "WriterHEPEVT.run_info": """Access the global GenRunInfo object.""",
    "WriterHEPEVT.vertices_positions_present": """get flag if vertex positions are available. The flag is deduced from m_options. If the m_options have the key "vertices_positions_are_absent" the result if false. True otherwise.

set flag if vertex positions are available. Effectively this adds or removes key "vertices_positions_are_absent" to/from the m_options.""",
    "WriterHEPEVT.write_event": """Write event to file.

Parameters
----------
evt:
    Event to be serialized.""",
    "WriterHEPEVT.write_hepevt_event_header": """Write event header to file.""",
    "WriterHEPEVT.write_hepevt_particle": """Write particle to file.

Parameters
----------
index:
    Particle to be serializediflong:
    Format of record.""",
    "delta_eta": """Pseudorapidity separation.""",
    "delta_phi": """Signed azimuthal angle separation in [-pi, pi].""",
    "delta_r2_eta": """R_eta^2-distance separation dR^2 = dphi^2 + deta^2.""",
    "delta_r2_rap": """R_rap^2-distance separation dR^2 = dphi^2 + drap^2.""",
    "delta_r_eta": """R_eta-distance separation dR = sqrt(dphi^2 + deta^2).""",
    "delta_r_rap": """R-rap-distance separation dR = sqrt(dphi^2 + drap^2).""",
    "delta_rap": """Rapidity separation.""",
}



from typing import Any, Tuple, Generator, Dict


def clear(self: Any) -> None:
    for key in list(self):
        del self[key]


def items(
    self: Any,
) -> Generator[Tuple[str, Any], None, None]:
    for name in self:
        yield name, self[name]


def eq(self: Any, other: Dict[str, Any]) -> bool:
    if len(self) != len(other):
        return False
    for k, v in self.items():
        if k not in other:
            return False
        if v != other[k]:
            return False
    return True


def repr(self: Any) -> str:
    s = f"<{self.__class__.__name__}>{{"
    first = True
    for k, v in self.items():
        if not first:
            s += ", "
        first = False
        s += f"{k!r}: {v!r}"
    s += "}"
    return s


def install() -> None:
    from pyhepmc._core import AttributesView, RunInfoAttributesView

    AttributesView.clear = clear
    AttributesView.items = items
    AttributesView.__eq__ = eq
    AttributesView.__repr__ = repr

    RunInfoAttributesView.clear = clear
    RunInfoAttributesView.items = items
    RunInfoAttributesView.__eq__ = eq
    RunInfoAttributesView.__repr__ = repr



from pyhepmc._deprecated import deprecated
import pytest
import numpy as np


class Foo:
    @deprecated("use baz")
    def bar(self):
        return 0


def test_deprecated():
    foo = Foo()

    with pytest.raises(ValueError, match="reason.*required"):

        @deprecated
        def bad(self):
            return 0

    with pytest.warns(np.VisibleDeprecationWarning, match="bar is deprecated: use baz"):
        foo.bar()



import pytest
import pyhepmc as hep
from numpy.testing import assert_equal


def make_evt():
    r"""
    In this example we will generate the following event by hand

        name status pdg_id      parent Px       Py     Pz       Energy      Mass
     1  !p+!    1   2212        0,0    0.000    0.000  7000.000 7000.000    0.938
     2  !He4!   2   1000020040  0,0    0.000    0.000 -7000.000 7000.000    3.756
    =============================================================================
     3  !d!     3      1        1,1    0.750   -1.569    32.191   32.238    0.000
     4  !u~!    4     -2        2,2   -3.047  -19.000   -54.629   57.920    0.000
     5  !W-!    5    -24        3,4    1.517   -20.68   -20.605   85.925   80.799
     6  !gamma! 6     22        3,4   -3.813    0.113    -1.833    4.233    0.000
     7  !d!     7      1        5,5   -2.445   28.816     6.082   29.552    0.010
     8  !u~!    8     -2        5,5    3.962  -49.498   -26.687   56.373    0.006

    The corresponding graph looks like this

                           p7
     p1                   /
       \v1__p3      p5---v4
             \_v3_/       \
             /    \        p8
        v2__p4     \
       /            p6
     p2

    Finally, the event gets a weight.
    """
    evt = hep.GenEvent(hep.Units.GEV, hep.Units.MM)
    evt.event_number = 1

    #                     px   py   pz      e        pdgid status
    p1 = hep.GenParticle((0.0, 0.0, 7000.0, 7000.0), 2212, 1)
    p1.generated_mass = 0.938
    p2 = hep.GenParticle((0.0, 0.0, -7000.0, 7000.0), 1000020040, 2)
    p2.generated_mass = 3.756
    p3 = hep.GenParticle((0.750, -1.569, 32.191, 32.238), 1, 3)
    p3.generated_mass = 0
    p4 = hep.GenParticle((-3.047, -19.0, -54.629, 57.920), -2, 4)
    p4.generated_mass = 0
    p5 = hep.GenParticle((1.517, -20.68, -20.605, 85.925), -24, 5)
    p5.generated_mass = 80.799
    p6 = hep.GenParticle((-3.813, 0.113, -1.833, 4.233), 22, 6)
    p6.generated_mass = 0
    p7 = hep.GenParticle((-2.445, 28.816, 6.082, 29.552), 1, 7)
    p7.generated_mass = 0.01
    p8 = hep.GenParticle((3.962, -49.498, -26.687, 56.373), -2, 8)
    p8.generated_mass = 0.006
    evt.add_particle(p1)
    evt.add_particle(p2)
    evt.add_particle(p3)
    evt.add_particle(p4)
    evt.add_particle(p5)
    evt.add_particle(p6)
    evt.add_particle(p7)
    evt.add_particle(p8)

    # make sure vertex is not optimized away by WriterAscii
    v1 = hep.GenVertex((1.0, 1.0, 1.0, 1.0))
    v1.add_particle_in(p1)
    v1.add_particle_out(p3)
    evt.add_vertex(v1)

    # make sure vertex is not optimized away by WriterAscii
    v2 = hep.GenVertex((2.0, 2.0, 2.0, 2.0))
    v2.add_particle_in(p2)
    v2.add_particle_out(p4)
    evt.add_vertex(v2)

    # make sure vertex is not optimized away by WriterAscii
    v3 = hep.GenVertex((3.0, 3.0, 3.0, 3.0))
    v3.add_particle_in(p3)
    v3.add_particle_in(p4)
    v3.add_particle_out(p5)
    v3.add_particle_out(p6)
    evt.add_vertex(v3)

    # make sure vertex is not optimized away by WriterAscii
    v4 = hep.GenVertex((4.0, 4.0, 4.0, 4.0))
    v4.add_particle_in(p5)
    v4.add_particle_out(p7)
    v4.add_particle_out(p8)
    evt.add_vertex(v4)

    evt.weights = [1.0]

    ti = hep.GenRunInfo.ToolInfo("dummy", "1.0", "dummy generator")
    ri = hep.GenRunInfo()
    ri.tools = [ti]
    ri.weight_names = ["0"]
    evt.run_info = ri

    return evt


@pytest.fixture()
def evt():
    return make_evt()


def test_GenHeavyIon():
    hi = hep.GenHeavyIon()
    assert hi == hep.GenHeavyIon()
    hi.Ncoll_hard = 3
    assert hi != hep.GenHeavyIon()


def test_GenParticle_as_attribute(evt):
    p = evt.particles[3]
    evt.attributes["foo"] = p
    assert evt.attributes == {"foo": p}


def test_GenCrossSection():
    cs = hep.GenCrossSection()
    cs.set_cross_section(1.2, 0.2, 3, 10)
    assert cs.event is None
    cs.set_xsec(0, 1.2)
    with pytest.raises(KeyError):
        cs.set_xsec_err("foo", 0.2)
    with pytest.raises(KeyError):
        cs.xsec("foo")
    assert cs.xsec(0) == 1.2
    assert cs.xsec_err(0) == 0.2
    assert cs.xsec() == 1.2
    with pytest.raises(IndexError):
        assert cs.xsec(1)

    ri = hep.GenRunInfo()
    ri.weight_names = ("foo", "bar")  # optional
    evt = hep.GenEvent()
    evt.run_info = ri
    evt.weights = [1.0, 2.0]
    evt.cross_section = cs
    assert evt.cross_section.event is evt
    cs = evt.cross_section
    cs.set_xsec("foo", 1.3)
    cs.set_xsec("bar", 2.3)
    assert cs.xsec("foo") == 1.3
    assert cs.xsec("bar") == 2.3
    assert cs.xsec(0) == 1.3
    assert cs.xsec(1) == 2.3
    with pytest.raises(IndexError):
        cs.xsec(2)
    with pytest.raises(KeyError):
        cs.xsec("baz")
    with pytest.raises(KeyError):
        cs.xsec_err("baz")


def test_attributes_1():
    ri = hep.GenRunInfo()
    att = ri.attributes
    assert att == {}
    assert len(att) == 0
    assert repr(att) == r"<RunInfoAttributesView>{}"
    att["foo"] = 1
    att["bar"] = "xy"
    att["baz"] = True
    assert att["foo"] == 1
    assert att["bar"] == "xy"
    assert att["baz"] is True
    assert att != {}
    assert att != {"foo": 2, "bar": "xy", "baz": True}
    assert att != {"foo": 1, "bra": "xy", "baz": True}
    with pytest.raises(KeyError):
        att["xyz"]
    assert len(att) == 3
    assert att == {"baz": True, "foo": 1, "bar": "xy"}
    # AttributeMapView has sorted keys
    assert repr(att) == r"<RunInfoAttributesView>{'bar': 'xy', 'baz': True, 'foo': 1}"

    del att["bar"]
    assert len(att) == 2
    assert att == {"baz": True, "foo": 1}

    keys = [k for k in att]
    assert keys == ["baz", "foo"]

    assert len(ri.attributes) == 2
    assert ri.attributes == att
    att.clear()
    assert len(att) == 0
    assert att == {}

    del ri
    # att should keep GenRunInfo alive through shared_ptr
    assert len(att) == 0


def test_attributes_2(evt):
    att = evt.attributes
    assert att == {}
    assert len(att) == 0
    assert repr(att) == r"<AttributesView>{}"
    att["foo"] = 1
    att["bar"] = "xy"
    att["baz"] = True
    assert att["foo"] == 1
    assert att["bar"] == "xy"
    assert att["baz"] is True
    with pytest.raises(KeyError):
        att["xyz"]
    assert len(att) == 3
    assert att == {"baz": True, "foo": 1, "bar": "xy"}
    # AttributeMapView has sorted keys
    assert repr(att) == r"<AttributesView>{'bar': 'xy', 'baz': True, 'foo': 1}"

    del att["bar"]
    assert len(att) == 2
    assert att == {"baz": True, "foo": 1}

    keys = [k for k in att]
    assert keys == ["baz", "foo"]

    assert len(evt.attributes) == 2
    assert evt.attributes == att
    att.clear()
    assert len(att) == 0
    assert att == {}


@pytest.mark.parametrize(
    "value",
    [
        True,
        2,
        1.5,
        "baz",
        [1, 2],
        ["foo", "bar"],
        [True, False],
        hep.GenCrossSection(),
        hep.GenHeavyIon(),
        hep.GenPdfInfo(),
        hep.HEPRUPAttribute(),
        hep.HEPEUPAttribute(),
    ],
)
def test_attributes_3(evt, value):
    if isinstance(value, hep.GenCrossSection):
        value.set_cross_section(1.2, 0.2, 3, 10)
    p1 = evt.particles[0]
    assert p1.id == 1
    assert p1.attributes == {}
    p1.attributes = {"foo": value}
    assert p1.attributes == {"foo": value}
    p1.attributes = {"bar": value}
    assert p1.attributes == {"bar": value}
    p1.attributes = {}
    assert p1.attributes == {}

    v1 = evt.vertices[0]
    assert v1.id != p1.id
    assert v1.attributes == {}
    v1.attributes = {"foo": value}
    assert v1.attributes == {"foo": value}
    v1.attributes = {"bar": value}
    assert v1.attributes == {"bar": value}
    v1.attributes = {}
    assert v1.attributes == {}


def test_FourVector():
    a = hep.FourVector(1, 2, 3, 4)
    b = hep.FourVector([1, 2, 3, 4])
    assert a == b

    with pytest.raises(ValueError):
        hep.FourVector([1, 2, 3])

    with pytest.raises(ValueError):
        hep.FourVector([1, 2, 3, 4, 5])


def test_GenPdfInfo(evt):
    pi = hep.GenPdfInfo()
    pi.parton_id1 = 211
    pi.parton_id2 = 2212
    pi.x1 = 0.5
    pi.x2 = 0.3
    assert pi.parton_id1 == 211
    assert pi.parton_id2 == 2212
    assert pi.x1 == 0.5
    assert pi.x2 == 0.3
    assert pi.scale == 0.0
    pi.scale = 1.2
    assert pi.scale == 1.2
    assert evt.pdf_info is None
    evt.pdf_info = pi
    assert evt.pdf_info == pi


def test_GenEvent(evt):
    for i, p in enumerate(evt.particles):
        assert p.status == i + 1

    p1, p2, p3, p4, *rest = evt.particles
    assert p1.pid == 2212
    assert p1.momentum == (0, 0, 7000, 7000)
    assert p1.generated_mass == 0.938
    assert p2.pid == 1000020040
    assert p2.generated_mass == 3.756
    assert p2.momentum == (0, 0, -7000, 7000)
    assert p3.parents == [p1]
    assert p3.momentum == (0.750, -1.569, 32.191, 32.238)
    assert p4.parents == [p2]
    assert p4.momentum == (-3.047, -19.0, -54.629, 57.920)


def test_GenEvent_generated_mass():
    p = hep.GenParticle((0.0, 0.0, 3.0, 5.0), 2212, 1)
    assert p.is_generated_mass_set() is False
    assert p.generated_mass == 4.0
    p.generated_mass = 2.3
    assert p.is_generated_mass_set() is True
    assert p.generated_mass == 2.3
    p.unset_generated_mass()
    assert p.generated_mass == 4.0


@pytest.mark.parametrize("use_parent", (True, False))
def test_GenEvent_from_hepevt(use_parent, evt):
    status = [p.status for p in evt.particles]
    pid = [p.pid for p in evt.particles]
    px = [p.momentum[0] for p in evt.particles]
    py = [p.momentum[1] for p in evt.particles]
    pz = [p.momentum[2] for p in evt.particles]
    en = [p.momentum[3] for p in evt.particles]
    m = [p.generated_mass for p in evt.particles]
    vx = vy = vz = vt = [0, 0, 1, 2, 3, 3, 4, 4]

    assert len(m) == 8

    #                           p7
    #     p1                   /
    #       \v1__p3      p5---v4
    #             \_v3_/       \
    #             /    \        p8
    #        v2__p4     \
    #       /            p6
    #     p2

    parents = [(0, 0), (0, 0), (1, 1), (2, 2), (3, 4), (3, 4), (5, 5), (5, 5)]
    children = [(3, 0), (4, 0), (5, 6), (5, 6), (7, 8), (0, 0), (0, 0), (0, 0)]

    ev = hep.GenEvent()
    if use_parent:
        ev.from_hepevt(
            evt.event_number,
            px,
            py,
            pz,
            en,
            m,
            pid,
            status,
            parents,
            None,
            vx,
            vy,
            vz,
            vt,
        )
    else:
        ev.from_hepevt(
            evt.event_number,
            px,
            py,
            pz,
            en,
            m,
            pid,
            status,
            None,
            children,
            vx,
            vy,
            vz,
            vt,
        )

    # cannot be taken from HepEvt record, but is set for evt
    ev.weights = [1.0]
    ev.run_info = evt.run_info

    assert len(ev.particles) == len(evt.particles)
    assert ev.particles == evt.particles
    assert len(ev.vertices) == len(evt.vertices)
    assert hep.equal_vertex_sets(ev.vertices, evt.vertices)
    assert ev == evt


def test_sequence_access():
    evt = hep.GenEvent()
    evt.add_particle(hep.GenParticle())
    evt.particles[0].momentum = (1, 2, 3, 4)
    evt.particles[0].pid = 5
    evt.add_vertex(hep.GenVertex())
    evt.vertices[0].position = (1, 2, 3, 4)
    assert len(evt.particles) == 1
    assert evt.particles[0].id == 1
    assert evt.particles[0].pid == 5
    assert evt.particles[0].momentum == (1, 2, 3, 4)
    assert len(evt.vertices) == 1
    assert evt.vertices[0].id == -1
    assert evt.vertices[0].position == (1, 2, 3, 4)
    assert repr(evt) == (
        "<GenEvent momentum_unit=1, length_unit=0, event_number=0, "
        "particles=1, vertices=1, run_info=None>"
    )


def test_weights():
    evt = hep.GenEvent()
    assert evt.weights == []
    with pytest.raises(RuntimeError, match=".*requires the event to have a GenRunInfo"):
        evt.weight_names
    evt.run_info = hep.GenRunInfo()
    evt.run_info.weight_names = ["a", "b"]
    evt.weights = [2, 3]
    assert evt.weight(1) == 3
    assert evt.weight("a") == 2
    with pytest.raises(RuntimeError, match="no weight with given name"):
        evt.weight("foo")
    with pytest.raises(IndexError):
        evt.weight(2)
    with pytest.raises(RuntimeError, match="no weight with given name"):
        evt.set_weight("c", 4)


def test_GenParticle_repr():
    p = hep.GenParticle([1, 2, 3, 4], pid=211, status=2)
    assert repr(p) == "GenParticle(FourVector(1, 2, 3, 4), pid=211, status=2)"


def test_GenVertex_repr():
    p = hep.GenVertex([1, 2, 3, 4])
    assert repr(p) == "GenVertex(FourVector(1, 2, 3, 4))"


def test_GenEventData(evt):
    ed = hep.GenEventData()
    evt.write_data(ed)

    assert len(ed.vertices) == 4
    a = ed.vertices
    assert_equal(a["status"], [v.status for v in evt.vertices])
    assert_equal(a["x"], [v.position.x for v in evt.vertices])
    assert_equal(a["z"], [v.position.z for v in evt.vertices])

    assert len(ed.particles) == 8
    b = ed.particles
    assert_equal(b["status"], [p.status for p in evt.particles])
    assert_equal(b["pid"], [p.pid for p in evt.particles])
    assert_equal(b["px"], [p.momentum.px for p in evt.particles])
    assert_equal(b["e"], [p.momentum.e for p in evt.particles])

    a["status"] = 2
    assert ed.vertices[0]["status"] == 2
    ed.particles["mass"] = 123
    assert ed.particles[0]["mass"] == 123

    evt2 = hep.GenEvent()
    evt2.read_data(ed)
    assert_equal([v.status for v in evt2.vertices], [2] * 4)
    assert_equal([p.generated_mass for p in evt2.particles], [123] * 8)


def test_numpy_api(evt):
    pids = evt.numpy.particles.pid
    assert_equal(pids, [p.pid for p in evt.particles])
    x = evt.numpy.vertices.x
    assert_equal(x, [v.position.x for v in evt.vertices])



"""
Visualization for GenEvent.
"""
from __future__ import annotations
from pyhepmc._graphviz import Digraph
from pyhepmc._prettify import db as prettify
from pyhepmc import Units, GenEvent
import numpy as np
import os
from pathlib import PurePath
from typing import BinaryIO, Union, Set, Any, Optional, Tuple


__all__ = ("to_dot", "savefig", "SUPPORTED_FORMATS")


def _supported_formats() -> Set[str]:
    import subprocess as subp

    try:
        r = subp.run(["dot", "-T12345679"], stderr=subp.PIPE)
    except FileNotFoundError:
        return set()
    s = r.stderr.decode("ascii").strip()
    match = "Use one of: "
    idx = s.index(match)
    assert idx > 0
    idx += len(match)
    # Some formats (eps, pov, ...) cannot handle utf-8 characters
    # GD Warning: GD image support has been disabled
    unsupported = {
        "eps",
        "cgimage",
        "pov",
        "ico",
        "sgi",
        "ismap",
        "imap",
        "imap_np",
        "pct",
        "pict",
        "ismap",
        "gd",
        "gd2",
        "json0",
        "x11",
        "xlib",
        "gtk",
        "tk",
        "vdx",
        "icns",
        "metafile",
    }
    formats = set(s[idx:].split()) - unsupported
    return formats


SUPPORTED_FORMATS = _supported_formats()


def to_dot(
    evt: GenEvent,
    *,
    size: Optional[Tuple[int, int]] = None,
    color_hadron: str = "black",
    color_lepton_or_boson: str = "goldenrod",
    color_quark_or_gluon: str = "darkred",
    color_internal: str = "dodgerblue",
    color_invalid: str = "gainsboro",
    color_special: str = "green",
) -> Digraph:
    """
    Convert GenEvent to Digraph.

    This converts the GenEvent into a graphviz Digraph in the DOT language,
    which in turn can be rendered in SVG format and displayed in Jupyter
    notebooks.

    Parameters
    ----------
    evt : GenEvent
        The event to convert.
    size : (int, int) or None, optional
        Maximum size of the graph in inches.
    color_hadron : str, optional
        Color (HTML color specification) for hadrons.
    color_lepton_or_boson: str, optional
        Color (HTML color specification) for leptons or bosons (gamma, W+/-, Z, H).
    color_quark_or_gluon: str, optional
        Color (HTML color specification) for quarks, diquarks, and gluons.
    color_internal: str, optional
        Color (HTML color specification) for generator-internal particles
        (e.g. strings, clusters, ...).
    color_special: str, optional
        Color (HTML color specification) for any valid particle which does not
        fit into the other categories (e.g. BSM particles).
    color_invalid: str, optional
        Color (HTML color specification) for invalid particles (e.g. PID==0).
    """
    if evt.run_info:
        name = "\n".join(f"{t.name}-{t.version}" for t in evt.run_info.tools)
        name += "\n"
    else:
        name = ""
    name += f"event number = {evt.event_number}"

    d = Digraph(name=name)
    d.node_attr["shape"] = "point"
    d.graph_attr["rankdir"] = "LR"
    if size is None:
        size_s = "8,100"
    else:
        size_s = f"{size[0]},{size[1]}"
    d.graph_attr["size"] = size_s

    GeV = 1 if evt.momentum_unit == Units.GEV else 1e3

    for v in evt.vertices:
        d.node(f"{v.id}", tooltip=f"{v.position} mm\nstatus = {v.status}")

    enmax = max(p.momentum.e / GeV for p in evt.particles)
    for p in evt.particles:
        en = p.momentum.e / GeV
        penwidth = str(max(0.5, 4 * en / enmax))

        unit = "GeV"
        if en > 1e2:
            en /= 1e3
            unit = "TeV"
        elif en < 1e-2:
            en *= 1e3
            unit = "MeV"

        style = "solid"
        color = "black"  # default color

        pname = prettify.get(p.pid, None)
        tooltip = [
            f"PDGID = {p.pid}",
            f"{p.momentum} GeV",
            f"m = {p.generated_mass:.4g} GeV",
            f"status = {p.status}",
        ]

        try:
            from particle import PDGID, Particle, ParticleNotFound, InvalidParticle

            pdgid = PDGID(p.pid)

            if pdgid.is_hadron or pdgid.is_nucleus:
                if pdgid.is_nucleus:
                    tooltip.append(f"A,Z = {pdgid.A},{pdgid.Z}")
                color = color_hadron
            elif pdgid == 21 or pdgid.is_quark or pdgid.is_diquark:
                color = color_quark_or_gluon
            elif pdgid.is_lepton or pdgid.is_gauge_boson_or_higgs:
                color = color_lepton_or_boson
            elif not pdgid.is_valid:
                color = color_invalid
                pname = f"Invalid({p.pid})"
            elif pdgid.is_generator_specific:
                color = color_internal
                pname = f"Internal({p.pid})"
            else:
                color = color_special

            try:
                pdb = Particle.from_pdgid(p.pid)
                if pdb.quarks:
                    tooltip.append(f"quarks = {pdb.quarks}")
                tooltip.append(f"Q = {pdb.charge:.3g}")
                if pdb.ctau and np.isfinite(pdb.ctau):
                    tooltip.append(f"cτ = {pdb.ctau:.3g} mm")
                if pdb.charge == 0:
                    style = "dashed"
            except (ParticleNotFound, InvalidParticle):
                pass
        except ModuleNotFoundError:
            pass

        if pname is None:
            pname = f"PDGID({p.pid})"

        label = f"{pname} {en:.2g} {unit}"
        tooltip = "\n".join(tooltip)

        # do not connect initial particles to root vertex
        if p.production_vertex and p.production_vertex.id != 0:
            vin = f"{p.production_vertex.id}"
        else:
            vin = f"in_{p.id}"
            d.node(vin, style="invis")

        if p.end_vertex:
            vout = f"{p.end_vertex.id}"
        else:
            vout = f"out_{p.id}"
            d.node(vout, style="invis")

        if not p.production_vertex or p.production_vertex.id == 0 and p.end_vertex:
            # initial particle which is not also final state is beam particle
            edir = "forward"
        elif p.status == 1:  # final state
            edir = "forward"
        else:  # intermediate
            edir = "none"

        d.edge(
            vin,
            vout,
            dir=edir,
            label=label,
            color=color,
            style=style,
            penwidth=penwidth,
            tooltip=tooltip,
            labeltooltip=tooltip,
        )

    return d


def savefig(
    event: Union[GenEvent, Digraph],
    fname: Union[str, BinaryIO],
    *,
    format: str = None,
    **kwargs: Any,
) -> None:
    """
    Save event as an image.

    The SVG format is recommended, because it contains tooltips with extra information.

    Supported formats: {SUPPORTED_FORMATS}.

    Parameters
    ----------
    event : GenEvent or Digraph
        The event to be saved.
    fname : path-like or file-like
        Path or file-like handle to which the image is written.
    format : str, optional
        Output format.
    **kwargs :
        Other arguments are forwarded to pyhepmc.view.to_dot.
    """
    if isinstance(fname, (str, os.PathLike)):
        p = PurePath(fname)
        if format is None:
            format = "".join(p.suffixes)[1:]
        if format in SUPPORTED_FORMATS:
            with open(p, "wb") as fo:
                savefig(event, fo, format=format, **kwargs)
        # unknown format raises exception in nested call
        with open(os.devnull, "wb") as fo:
            savefig(event, fo, format=format, **kwargs)
        return

    # if we arrive here, fname is a file-like object

    if format is None:
        raise ValueError("When using file-like object, keyword 'format' must be set")

    if format not in SUPPORTED_FORMATS:
        raise ValueError(
            f"Format {format!r} not supported (supported formats: "
            f"{', '.join(SUPPORTED_FORMATS)})"
        )

    if isinstance(event, GenEvent):
        g = to_dot(event, **kwargs)
    else:
        if kwargs:
            import warnings

            warnings.warn(
                "Extra kwargs are ignored when passing DiGraph", RuntimeWarning
            )
        g = event

    s = g.pipe(format=format)
    assert isinstance(s, bytes)
    fname.write(s)


savefig.__doc__ = savefig.__doc__.format(  # type:ignore
    SUPPORTED_FORMATS=", ".join(SUPPORTED_FORMATS)
)



import sys
from pathlib import Path
from setuptools import setup
from distutils.command.sdist import sdist

cdir = Path(__file__).parent.absolute()

sys.path.append(str(cdir))
from cmake_ext import CMakeExtension, CMakeBuild  # noqa: E402
from setup_utils import merge_license_files  # noqa: E402


class sdist_mod(sdist):
    def run(self):
        with merge_license_files():
            sdist.run(self)


setup(
    zip_safe=False,
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
    ext_modules=[CMakeExtension("pyhepmc._core")],
    cmdclass={"build_ext": CMakeBuild, "sdist": sdist_mod},
)



from contextlib import contextmanager


@contextmanager
def merge_license_files():
    """
    Merge LICENSE.txt and LICENSES_bundled.txt for sdist creation.

    This follows the approach of Scipy and is to keep LICENSE in repo as an
    exact BSD 3-clause, to make GitHub state correctly how pyhepmc is licensed.
    """

    l1 = "LICENSE"
    l2 = "LICENSES_bundled"

    with open(l1, "r") as f:
        content1 = f.read()

    with open(l2, "r") as f:
        content2 = f.read()

    with open(l1, "w") as f:
        f.write(content1 + "\n\n" + content2)

    yield

    with open(l1, "w") as f:
        f.write(content1)



# flake8: noqa
from ._autodoc import autodoc as doc

override = {
    "Units": "Units in which HepMC3 stores momentum and position.",
    "GenParticle.generated_mass": """Get or set generated mass.

This function will return mass as set by a generator/tool. If not set, it will return momentum().m().""",
    "GenParticle.parents": """Convenience access to immediate incoming particles via production vertex.

Less efficient than via the vertex since return must be by value (in case there is no vertex).
""",
    "equal_vertex_sets": "Whether two sets of vertices are equal. The vertices can have arbitrary order and their id() is ignored in the comparison. The sets are considered equal, if each vertex in one set compares equal to another vertex in the other set.",
    "equal_particle_sets": "Whether two sets of particles are equal. Algorithm is analog to :func:`equal_vertex_sets`.",
    "GenCrossSection": """Stores additional information about cross-section.

This is an example of event attribute used to store cross-section information.

This class is meant to be used to pass, on an event by event basis, the current best guess of the total cross section. it is expected that the final cross section will be stored elsewhere.

Several cross sections and related info can be included in case of runs with multiple weights.

The units of cross-sections are expected to be pb.""",
    "GenParticle.id": "Number which uniquely identifies this particle in the current event. This is not a particle ID, see :attr:`pid` for the PDG ID. The id is a persistent pointer to this particle.",
    "GenVertex.id": "Number which uniquely identifies this vertex in the current event. The id is a persistent pointer to this vertex.",
    "GenParticle.pid": "PDG ID of particle",
    "GenParticle.abs_pid": "Equivalent to abs(genparticle.pid)",
    "GenParticle.is_generated_mass_set": "See :attr:`generated_mass`.",
    "GenParticle.unset_generated_mass": "Declare that generated mass is not set. See :attr:`generated_mass`.",
    "GenEvent.from_hepevt": """
    Convert HEPEVT record to GenEvent.

    GenEvent is cleared and then filled with the information provided in the argument arrays. Recovering the particle history only requires either parents or children.
    If both are not None, parents are used.

    All particle arrays must have equal length.

    Parameters
    ----------
    event_number: int
        Event number, starting with zero.
    px : array-like
        X-component of momentum of particles in GeV.
    py : array-like
        Y-component of momentum of particles in GeV.
    pz : array-like
        Z-component of momentum of particles in GeV.
    en : array-like
        Energy of particles in GeV.
    m : array-like
        Generated mass of particles in GeV.
    pid : array-like
        PDG IDs of particles.
    status: array-like
        Status codes of particles.
    parents: array-like or None, optional
        Array of int with shape (N, 2). Start and stop index (inclusive) in the record for the parents of each particle in Fortran numbering (starting with 1). No parents for a particle are indicated by the pair (0, 0) or (-1, -1). Single parents are indicated either by (N, 0) or (N, N) with N > 0.
    children: array-like or None, optional
        Like parents but for the children of this particle.
    vx : array-like or None, optional
        X-component of location of production vertex of particles in mm.
    vy : array-like or None, optional
        Y-component of location of production vertex of particles in mm.
    vz : array-like or None, optional
        Z-component of location of production vertex of particles in mm.
    vt : array-like or None, optional
        Time (ct) of production vertex of particles in mm.
    """,
    "GenEvent.weight": """Get event weight accessed by index (or the canonical/first one if there is no argument) or name.

    Access by weight name requires a :class:`GenRunInfo` attached to the event, otherwise this will throw an exception.

    It's the user's responsibility to ensure that the index or name exists!""",
    "GenEvent.set_weight": """Set event weight accessed by index (or the canonical/first one if there is no argument) or name.

    Access by weight name requires a :class:`GenRunInfo` attached to the event, otherwise this will throw an exception.

    It's the user's responsibility to ensure that the index or name exists!""",
    "GenEvent.vertices": "",
    "GenEvent.weight_names": """Get event weight names.

    Access requires a :class:`GenRunInfo` attached to the event, otherwise this will throw an exception.""",
    "GenEvent.weights": """Access event weight values as a sequence.

    Can be used to set multiple weights by assigning a sequence.
    """,
    "GenEvent.beams": """Access beam particles as a sequence.

    HepMC3 considers every particle as a beam particles, which does not have another ancestor particle. Therefore, this is a read-only property.
    """,
    "GenEvent.cross_section": """
    Access to the :class:`GenCrossSection`.

    If this attribute is not set, returns None.
    """,
    "GenEvent.heavy_ion": """
    Access to the :class:`GenHeavyIon`.

    If this attribute is not set, returns None.
    """,
    "GenEvent.heavy_ion": """
    Access to the :class:`GenHeavyIon`.

    If this attribute is not set, returns None.
    """,
    "GenEvent.run_info": """
    Access to the :class:`GenRunInfo`.

    If this attribute is not set, returns None.
    """,
    "GenEvent.pdf_info": """
    Access to the :class:`GenPdfInfo`.

    If this attribute is not set, returns None.
    """,
    "GenEvent.vertices": "Access list of vertices.",
    "attributes": """Access attributes with a dict-like view.

    It is possible to read and write attributes. Primitive C++ types (and vectors therefore) are converted from/to native Python types.
    """,
    "UnparsedAttribute": """Unparsed attribute after deserialization.

    HepMC3 does not serialize the type of attributes, therefore the correct
    type cannot be restored upon deserialization (this is a limition of the HepMC3 C++
    library and its serialization format). Use the :meth:`astype` method to parse the
    attribute into a concrete type; this has important side-effects, see method
    description.
    """,
    "UnparsedAttribute.astype": """
    Convert unparsed attribute to concrete type.

    If the conversion is successful, the unparsed attribute is replaced with the parsed
    attribute, so this method has to be called only once. If the conversion fails, a
    TypeError is raised.

    Parameters
    ----------
    pytype: type
        Type of the attribute. Allowed values: bool, int, float, str, GenParticle,
        GenPdfInfo, GenHeavyIon, GenCrossSection, HEPRUPAttribute, HEPEUPAttribute,
        typing.List[int], typing.List[float], typing.List[str]. In Python-3.9+,
        typing.List can be replaced by list.
    """,
}

doc.update(override)
